import os

from supriya import HeaderFormat, SampleFormat
from supriya.system import SupriyaObject


class Recorder(SupriyaObject):
    """
    A server recorder.

    ::

        >>> import os, tempfile, time
        >>> server = supriya.Server().boot()
        >>> with tempfile.TemporaryDirectory() as tempdir:
        ...     server.recorder.start(
        ...         file_path=os.path.join(tempdir, "example.aiff"),
        ...         channel_count=2,
        ...         header_format="AIFF",
        ...     )
        ...     time.sleep(1)
        ...     server.recorder.stop()
        ...

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_channel_count",
        "_current_channel_count",
        "_current_header_format",
        "_current_file_path",
        "_current_sample_format",
        "_header_format",
        "_is_recording",
        "_record_buffer",
        "_record_node",
        "_record_synthdef",
        "_sample_format",
        "_server",
    )

    ### INITIALIZER ###

    def __init__(self, server):
        self._server = server
        server_options = server.options
        # setup settings
        self._channel_count = server_options.output_bus_channel_count
        self._header_format = HeaderFormat.AIFF
        self._is_recording = False
        self._record_node = None
        self._record_buffer = None
        self._record_synthdef = None
        self._sample_format = SampleFormat.INT24
        # cache settings
        self._current_channel_count = self._channel_count
        self._current_file_path = None
        self._current_header_format = self._header_format
        self._current_sample_format = self._sample_format

    ### PRIVATE METHODS ###

    def _cache_properties(
        self, channel_count=None, header_format=None, sample_format=None
    ):
        if channel_count is None:
            channel_count = self.channel_count
        self._current_channel_count = channel_count
        if header_format is None:
            header_format = self.header_format
        self._current_header_format = HeaderFormat.from_expr(header_format)
        if sample_format is None:
            sample_format = self.sample_format
        self._current_sample_format = SampleFormat.from_expr(sample_format)

    def _get_file_path(self, file_path=None):
        if file_path is not None:
            file_path = os.path.abspath(os.path.expanduser(file_path))
        return file_path

    def _setup_buffer(self):
        import supriya.commands
        import supriya.realtime

        frame_count = 65536
        buffer_ = supriya.realtime.Buffer().allocate(
            self.server,
            frame_count=frame_count,
            channel_count=self.current_channel_count,
        )
        callback = supriya.commands.BufferWriteRequest(
            buffer_id=buffer_,
            file_path=self.current_file_path,
            frame_count=0,
            header_format=self.current_header_format,
            leave_open=True,
            sample_format=self.current_sample_format,
            starting_frame=0,
        )
        callback.communicate(server=self.server, sync=True)
        self._record_buffer = buffer_

    def _setup_node(self):
        import supriya.realtime

        synth = supriya.realtime.Synth(self.record_synthdef)
        synth.allocate(
            add_action=supriya.AddAction.ADD_TO_TAIL,
            target_node=self.server.root_node,
            node_id_is_permanent=True,
        )
        self._record_node = synth

    def _setup_synthdef(self):
        import supriya.synthdefs
        import supriya.ugens

        with supriya.synthdefs.SynthDefBuilder() as builder:
            source = supriya.ugens.In.ar(
                bus=0, channel_count=self.current_channel_count
            )
            buffer_id = int(self.record_buffer)
            supriya.ugens.DiskOut.ar(buffer_id=buffer_id, source=source)
        synthdef = builder.build()
        synthdef.allocate(server=self.server)
        self._record_synthdef = synthdef

    ### PUBLIC METHODS ###

    def pause(self):
        if self.is_recording:
            self.record_node.pause()
        else:
            raise Exception

    def prepare(
        self, file_path=None, channel_count=None, header_format=None, sample_format=None
    ):
        if self.is_recording:
            raise Exception
        self._cache_properties(
            channel_count=channel_count,
            header_format=header_format,
            sample_format=sample_format,
        )
        self._current_file_path = self._get_file_path(file_path)
        if self.current_file_path is None:
            raise Exception
        self._setup_buffer()
        self._setup_synthdef()

    def start(
        self, file_path=None, channel_count=None, header_format=None, sample_format=None
    ):
        if self.record_node is not None:
            raise Exception("Already recording.")
        if not self.record_buffer or not self.record_buffer.is_allocated:
            self.prepare(
                file_path=file_path,
                channel_count=channel_count,
                header_format=header_format,
                sample_format=sample_format,
            )
        self._setup_node()
        self._is_recording = True

    def stop(self):
        if not self.is_recording:
            raise Exception
        self.record_node.free()
        self.record_buffer.close()
        self.record_buffer.free()
        self._is_recording = False

    def unpause(self):
        if self.is_recording:
            self.record_node.unpause()
        else:
            raise Exception

    ### PUBLIC PROPERTIES ###

    @property
    def is_recording(self):
        return self._is_recording

    @property
    def record_buffer(self):
        return self._record_buffer

    @property
    def record_node(self):
        return self._record_node

    @property
    def record_synthdef(self):
        return self._record_synthdef

    @property
    def channel_count(self):
        return self._channel_count

    @channel_count.setter
    def channel_count(self, expr):
        channel_count = int(expr)
        assert 0 < channel_count
        self._channel_count = channel_count

    @property
    def sample_format(self):
        return self._sample_format

    @sample_format.setter
    def sample_format(self, expr):
        sample_format = SampleFormat.from_expr(expr)
        self._sample_format = sample_format

    @property
    def header_format(self):
        return self._header_format

    @header_format.setter
    def header_format(self, expr):
        header_format = HeaderFormat.from_expr(expr)
        self._header_format = header_format

    @property
    def current_channel_count(self):
        return self._current_channel_count

    @property
    def current_file_path(self):
        return self._current_file_path

    @property
    def current_sample_format(self):
        return self._current_sample_format

    @property
    def current_header_format(self):
        return self._current_header_format

    @property
    def server(self):
        return self._server



import enum
from typing import Dict, Optional, Tuple
from uuid import UUID

from uqbar.objects import get_repr, get_vars, new

from supriya.enums import AddAction, CalculationRate
from supriya.synthdefs import SynthDef
from supriya.utils import expand


class Priority(enum.IntEnum):
    START = 1
    STOP = 2


class Event:
    def __init__(self, *, delta=0.0):
        self.delta = delta

    def __eq__(self, expr):
        self_values = type(self), get_vars(self)
        try:
            expr_values = type(expr), get_vars(expr)
        except AttributeError:
            expr_values = type(expr), expr
        return self_values == expr_values

    def __repr__(self):
        return get_repr(self, multiline=False)

    def expand(self, offset):
        return [(offset, Priority.START, self)]

    def perform(
        self,
        provider,
        proxy_mapping,
        *,
        current_offset: float,
        notes_mapping: Dict[Tuple[UUID, int], float],
        priority: int,
    ):
        raise NotImplementedError


class BusAllocateEvent(Event):
    def __init__(self, id_, *, calculation_rate="audio", channel_count=1, delta=0.0):
        Event.__init__(self, delta=delta)
        self.id_ = id_
        self.calculation_rate = CalculationRate.from_expr(calculation_rate)
        self.channel_count = channel_count

    def perform(
        self,
        provider,
        proxy_mapping,
        *,
        current_offset: float,
        notes_mapping: Dict[Tuple[UUID, int], float],
        priority: int,
    ):
        proxy_mapping[self.id_] = provider.add_bus_group(
            calculation_rate=self.calculation_rate, channel_count=self.channel_count
        )


class BusFreeEvent(Event):
    def __init__(self, id_, *, delta=0.0):
        Event.__init__(self, delta=delta)
        self.id_ = id_

    def perform(
        self,
        provider,
        proxy_mapping,
        *,
        current_offset: float,
        notes_mapping: Dict[Tuple[UUID, int], float],
        priority: int,
    ):
        provider.free_bus_group(proxy_mapping.pop(self.id_))


class CompositeEvent(Event):
    def __init__(self, events, *, delta=0.0):
        Event.__init__(self, delta=delta)
        self.events = events

    def expand(self, offset):
        events = []
        for event in self.events:
            events.extend(event.expand(offset))
            offset += event.delta
        return events


class GroupAllocateEvent(Event):
    def __init__(
        self, id_, *, add_action=AddAction.ADD_TO_HEAD, delta=0.0, target_node=None
    ):
        Event.__init__(self, delta=delta)
        self.id_ = id_
        self.add_action = AddAction.from_expr(add_action)
        self.target_node = target_node

    def perform(
        self,
        provider,
        proxy_mapping,
        *,
        current_offset: float,
        notes_mapping: Dict[Tuple[UUID, int], float],
        priority: int,
    ):
        proxy_mapping[self.id_] = provider.add_group(
            add_action=self.add_action, target_node=proxy_mapping.get(self.target_node)
        )


class NodeFreeEvent(Event):
    def __init__(self, id_, *, delta=0.0):
        Event.__init__(self, delta=delta)
        self.id_ = id_

    def perform(
        self,
        provider,
        proxy_mapping,
        *,
        current_offset: float,
        notes_mapping: Dict[Tuple[UUID, int], float],
        priority: int,
    ):
        provider.free_node(proxy_mapping.pop(self.id_))


class NoteEvent(Event):
    def __init__(
        self,
        id_,
        *,
        add_action=AddAction.ADD_TO_HEAD,
        delta: float = 1.0,
        duration: float = 1.0,
        synthdef: Optional[SynthDef] = None,
        target_node: Optional[UUID] = None,
        **kwargs,
    ):
        Event.__init__(self, delta=delta)
        self.id_ = id_
        self.add_action = AddAction.from_expr(add_action)
        self.duration = duration
        self.synthdef = synthdef
        self.target_node = target_node
        self.kwargs = kwargs

    def expand(self, offset: float):
        starts, stops = [], []
        for i, proxy_mapping in enumerate(expand(self.kwargs)):
            event = type(self)(
                id_=(self.id_, i),
                add_action=self.add_action,
                duration=self.duration,
                synthdef=self.synthdef,
                target_node=self.target_node,
                **proxy_mapping,
            )
            start_offset = offset
            stop_offset = offset + self.calculate_duration()
            starts.append((start_offset, Priority.START, event))
            stops.append((stop_offset, Priority.STOP, event))
        return starts + stops

    def calculate_duration(self):
        return self.duration or 1.0

    def merge(self, event):
        _, _, kwargs = get_vars(event)
        return new(self, **kwargs)

    def perform(
        self,
        provider,
        proxy_mapping,
        *,
        current_offset: float,
        notes_mapping: Dict[Tuple[UUID, int], float],
        priority: int,
        **kwargs,
    ):
        if priority == Priority.START:
            # does a proxy exist?
            #    if yes, update settings
            #    if no, create proxy
            # update notes mapping with expected completion offset
            settings = self.kwargs.copy()
            for key, value in settings.items():
                if isinstance(value, UUID):
                    settings[key] = proxy_mapping[value]
            if self.id_ not in proxy_mapping:
                proxy_mapping[self.id_] = provider.add_synth(
                    add_action=self.add_action,
                    synthdef=self.synthdef,
                    target_node=proxy_mapping.get(self.target_node),
                    **settings,
                )
            else:
                proxy = proxy_mapping[self.id_]
                provider.set_node(proxy, **settings)
            notes_mapping[self.id_] = current_offset + self.calculate_duration()
        elif priority == Priority.STOP:
            # check notes mapping for expected completion offset
            # is expected completion <= current_offset? release
            # otherwise pass
            expected_completion = notes_mapping.get(self.id_)
            if expected_completion is None or expected_completion > current_offset:
                return
            notes_mapping.pop(self.id_)
            provider.free_node(proxy_mapping.pop(self.id_))


class NullEvent(Event):
    def expand(self, offset):
        return []


class SynthAllocateEvent(Event):
    def __init__(
        self,
        id_,
        *,
        add_action=AddAction.ADD_TO_HEAD,
        delta=0.0,
        synthdef=None,
        target_node=None,
        **kwargs,
    ):
        Event.__init__(self, delta=delta)
        self.id_ = id_
        self.add_action = AddAction.from_expr(add_action)
        self.synthdef = synthdef
        self.target_node = target_node
        self.kwargs = kwargs

    def perform(
        self,
        provider,
        proxy_mapping,
        *,
        current_offset: float,
        notes_mapping: Dict[Tuple[UUID, int], float],
        priority: int,
    ):
        settings = self.kwargs.copy()
        for key, value in settings.items():
            if isinstance(value, UUID):
                settings[key] = proxy_mapping[value]
        proxy_mapping[self.id_] = provider.add_synth(
            add_action=self.add_action,
            synthdef=self.synthdef,
            target_node=proxy_mapping.get(self.target_node),
            **settings,
        )



from queue import Empty, PriorityQueue
from threading import RLock

from .events import Event


class PatternPlayer:
    def __init__(self, pattern, provider, clock):
        self._pattern = pattern
        self._provider = provider
        self._clock = clock
        self._lock = RLock()
        self._queue = PriorityQueue()
        self._is_running = False
        self._is_stopping = False
        self._proxies_by_uuid = {}
        self._notes_by_uuid = {}

    def _clock_callback(self, context, *args, **kwargs):
        with self._lock:
            current_offset = None
            events = []
            while True:
                try:
                    offset, priority, index, event = self._queue.get(block=False)
                except Empty:
                    self._perform_events(
                        context.desired_moment.seconds, current_offset, events
                    )
                    self._is_running = False
                    return None
                except Exception:
                    return
                if offset == float("-inf"):
                    offset = context.desired_moment.offset
                delta = offset - context.desired_moment.offset
                if delta:
                    self._queue.put((offset, priority, index, event))
                    self._perform_events(
                        context.desired_moment.seconds, current_offset, events
                    )
                    return delta
                if not isinstance(event, Event):
                    if self._consume_iterator(offset):
                        return
                elif offset != current_offset:
                    self._perform_events(
                        context.desired_moment.seconds, current_offset, events
                    )
                    current_offset = offset
                    events = [(event, priority)]
                else:
                    events.append((event, priority))

    def _consume_iterator(self, current_offset):
        try:
            try:
                index, consumed_event = self._iterator.send(self._is_stopping)
            except TypeError:
                if self._is_stopping:
                    return True
                index, consumed_event = next(self._iterator)
            for subindex, (expanded_offset, priority, expanded_event) in enumerate(
                consumed_event.expand(current_offset)
            ):
                self._queue.put(
                    (expanded_offset, priority, (index, subindex), expanded_event)
                )
            self._queue.put(
                (current_offset + consumed_event.delta, 0, (index, 0), None)
            )
        except StopIteration:
            pass
        return False

    def _stop_callback(self, context, *args, **kwargs):
        with self._lock:
            # Do we need to rebuild the queue? Yes.
            # Do we need to free all playing notes? Yes.
            # How do we handle when there are already stop events in the queue? They'll be no-ops when performed.
            self._is_stopping = True
            self._clock.reschedule(
                self._clock_event_id, schedule_at=context.desired_moment.offset
            )
            self._reschedule_queue(context.desired_moment.offset)
            self._free_all_notes(context.desired_moment.seconds)

    def _enumerate(self, iterator):
        index = 0
        should_stop = False
        while True:
            try:
                should_stop = yield (index, iterator.send(should_stop)) or should_stop
            except TypeError:
                should_stop = yield (index, next(iterator))
            except StopIteration:
                return
            index += 1

    def _free_all_notes(self, current_seconds):
        if not self._notes_by_uuid:
            return
        with self._provider.at(current_seconds):
            while self._notes_by_uuid:
                uuid, _ = self._notes_by_uuid.popitem()
                proxy = self._proxies_by_uuid.pop(uuid)
                self._provider.free_node(proxy)

    def _perform_events(self, current_seconds, current_offset, events):
        if not events:
            return
        with self._provider.at(current_seconds):
            for event, priority in events:
                event.perform(
                    self._provider,
                    self._proxies_by_uuid,
                    current_offset=current_offset,
                    notes_mapping=self._notes_by_uuid,
                    priority=priority,
                )

    def _reschedule_queue(self, current_offset):
        events = []
        while not self._queue.empty():
            events.append(self._queue.get())
        if not events:
            return
        delta = events[0][0] - current_offset
        for event in events:
            offset, *rest = event
            self._queue.put((offset - delta, *rest))

    def play(self, quantization: str = None, at=None, until=None):
        with self._lock:
            if self._is_running:
                return
            self._iterator = self._enumerate(iter(self._pattern))
            self._queue.put((float("-inf"), 0, (0, 0), None))
            self._is_running = True
            self._is_stopping = False
        self._clock_event_id = self._clock.cue(
            self._clock_callback, event_type=3, quantization=quantization
        )
        if until:
            self._clock.schedule(self._stop_callback, event_type=2, schedule_at=until)
        if not self._clock.is_running:
            self._clock.start(initial_time=at)

    def stop(self, quantization: str = None):
        with self._lock:
            if not self._is_running or self._is_stopping:
                return
            self._clock.cue(
                self._stop_callback, event_type=2, quantization=quantization
            )



import collections

from uqbar.enums import IntEnumeration

from supriya import CalculationRate
from supriya.synthdefs import MultiOutUGen, UGen


class BeatTrack(MultiOutUGen):
    """
    Autocorrelation beat tracker.

    ::

        >>> source = supriya.ugens.SoundIn.ar(bus=0)
        >>> pv_chain = supriya.ugens.FFT(source=source)
        >>> beat_track = supriya.ugens.BeatTrack.kr(
        ...     pv_chain=pv_chain,
        ...     lock=0,
        ... )
        >>> beat_track
        UGenArray({4})

    """

    _default_channel_count = 4
    _has_settable_channel_count = False
    _ordered_input_names = collections.OrderedDict([("pv_chain", None), ("lock", 0.0)])
    _valid_calculation_rates = (CalculationRate.CONTROL,)


class BeatTrack2(MultiOutUGen):
    """
    A template-matching beat-tracker.

    ::

        >>> beat_track_2 = supriya.ugens.BeatTrack2.kr(
        ...     bus_index=0,
        ...     lock=False,
        ...     feature_count=4,
        ...     phase_accuracy=0.02,
        ...     weighting_scheme=-2.1,
        ...     window_size=2,
        ... )
        >>> beat_track_2
        UGenArray({6})

    """

    _default_channel_count = 6
    _has_settable_channel_count = False
    _ordered_input_names = collections.OrderedDict(
        [
            ("bus_index", 0.0),
            ("feature_count", None),
            ("window_size", 2),
            ("phase_accuracy", 0.02),
            ("lock", 0.0),
            ("weighting_scheme", -2.1),
        ]
    )
    _valid_calculation_rates = (CalculationRate.CONTROL,)


class KeyTrack(UGen):
    """
    A key tracker.

    ::

        >>> source = supriya.ugens.SoundIn.ar(bus=0)
        >>> pv_chain = supriya.ugens.FFT(source=source)
        >>> key_track = supriya.ugens.KeyTrack.kr(
        ...     pv_chain=pv_chain,
        ...     chroma_leak=0.5,
        ...     key_decay=2,
        ... )
        >>> key_track
        KeyTrack.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain", None), ("key_decay", 2), ("chroma_leak", 0.5)]
    )
    _valid_calculation_rates = (CalculationRate.CONTROL,)


class Loudness(UGen):
    """
    Extraction of instantaneous loudness in `sones`.

    ::

        >>> source = supriya.ugens.SoundIn.ar(bus=0)
        >>> pv_chain = supriya.ugens.FFT(source=source)
        >>> loudness = supriya.ugens.Loudness.kr(
        ...     pv_chain=pv_chain,
        ...     smask=0.25,
        ...     tmask=1,
        ... )
        >>> loudness
        Loudness.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain", None), ("smask", 0.25), ("tmask", 1)]
    )
    _valid_calculation_rates = (CalculationRate.CONTROL,)


class MFCC(MultiOutUGen):
    """
    Mel frequency cepstral coefficients.

    ::

        >>> source = supriya.ugens.SoundIn.ar(bus=0)
        >>> pv_chain = supriya.ugens.FFT(source=source)
        >>> mfcc = supriya.ugens.MFCC.kr(
        ...     pv_chain=pv_chain,
        ...     channel_count=13,
        ... )
        >>> mfcc
        UGenArray({13})

    """

    _default_channel_count = 1
    _has_settable_channel_count = True
    _ordered_input_names = collections.OrderedDict([("pv_chain", None)])
    _valid_calculation_rates = (CalculationRate.CONTROL,)


class Onsets(UGen):
    """
    An onset detector.

    ::

        >>> source = supriya.ugens.SoundIn.ar(bus=0)
        >>> pv_chain = supriya.ugens.FFT(source=source)
        >>> onsets = supriya.ugens.Onsets.kr(
        ...     pv_chain=pv_chain,
        ...     floor=0.1,
        ...     medianspan=11,
        ...     mingap=10,
        ...     odftype=supriya.ugens.Onsets.ODFType.RCOMPLEX,
        ...     rawodf=0,
        ...     relaxtime=1,
        ...     threshold=0.5,
        ...     whtype=1,
        ... )
        >>> onsets
        Onsets.kr()

    """

    class ODFType(IntEnumeration):
        POWER = 0
        MAGSUM = 1
        COMPLEX = 2
        RCOMPLEX = 3
        PHASE = 4
        WPHASE = 5
        MKL = 6

    _ordered_input_names = collections.OrderedDict(
        [
            ("pv_chain", None),
            ("threshold", 0.5),
            ("odftype", 3),
            ("relaxtime", 1),
            ("floor", 0.1),
            ("mingap", 10),
            ("medianspan", 11),
            ("whtype", 1),
            ("rawodf", 0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.CONTROL,)


class SpecCentroid(UGen):
    """
    A spectral centroid measure.

    ::

        >>> source = supriya.ugens.SoundIn.ar(bus=0)
        >>> pv_chain = supriya.ugens.FFT(source=source)
        >>> spec_centroid = supriya.ugens.SpecCentroid.kr(
        ...     pv_chain=pv_chain,
        ... )
        >>> spec_centroid
        SpecCentroid.kr()

    """

    _ordered_input_names = collections.OrderedDict([("pv_chain", None)])
    _valid_calculation_rates = (CalculationRate.CONTROL,)


class SpecFlatness(UGen):
    """
    A spectral flatness measure.

    ::

        >>> source = supriya.ugens.SoundIn.ar(bus=0)
        >>> pv_chain = supriya.ugens.FFT(source=source)
        >>> spec_flatness = supriya.ugens.SpecFlatness.kr(
        ...     pv_chain=pv_chain,
        ... )
        >>> spec_flatness
        SpecFlatness.kr()

    """

    _ordered_input_names = collections.OrderedDict([("pv_chain", None)])
    _valid_calculation_rates = (CalculationRate.CONTROL,)


class SpecPcile(UGen):
    """
    Find a percentile of FFT magnitude spectrum.

    ::

        >>> source = supriya.ugens.SoundIn.ar(bus=0)
        >>> pv_chain = supriya.ugens.FFT(source=source)
        >>> spec_pcile = supriya.ugens.SpecPcile.kr(
        ...     pv_chain=pv_chain,
        ...     fraction=0.5,
        ...     interpolate=0,
        ... )
        >>> spec_pcile
        SpecPcile.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("pv_chain", None), ("fraction", 0.5), ("interpolate", 0)]
    )
    _valid_calculation_rates = (CalculationRate.CONTROL,)



import queue


class EventQueue(queue.PriorityQueue):

    ### PRIVATE METHODS ###

    def _init(self, maxsize):
        self.queue = []
        self.items = {}

    def _put(self, item):
        entry = [item, True]
        if item in self.items:
            self.items[item][-1] = False
        self.items[item] = entry
        super()._put(entry)

    def _get(self):
        while self.queue:
            item, active = super()._get()
            if active:
                del self.items[item]
                return item
        raise queue.Empty

    ### PUBLIC METHODS ###

    def clear(self):
        with self.mutex:
            self._init(None)

    def peek(self):
        with self.mutex:
            item = self._get()
            entry = [item, True]
            self.items[item] = entry
            super()._put(entry)
        return item

    def remove(self, item):
        with self.mutex:
            entry = self.items.pop(item, None)
            if entry is not None:
                entry[-1] = False



from .eventpatterns import ChainPattern, EventPattern, MonoEventPattern, UpdatePattern
from .events import (
    BusAllocateEvent,
    BusFreeEvent,
    CompositeEvent,
    GroupAllocateEvent,
    NodeFreeEvent,
    NoteEvent,
    NullEvent,
    SynthAllocateEvent,
)
from .noise import ChoicePattern, RandomPattern, ShufflePattern
from .patterns import (
    BinaryOpPattern,
    Pattern,
    SeedPattern,
    SequencePattern,
    UnaryOpPattern,
)
from .players import PatternPlayer
from .sequences import GatePattern, RepeatPattern, RestartPattern, StutterPattern
from .structure import BusPattern, FxPattern, GroupPattern, ParallelPattern

__all__ = [
    "BinaryOpPattern",
    "BusAllocateEvent",
    "BusFreeEvent",
    "BusPattern",
    "ChainPattern",
    "ChoicePattern",
    "CompositeEvent",
    "EventPattern",
    "FxPattern",
    "GatePattern",
    "GroupAllocateEvent",
    "GroupPattern",
    "MonoEventPattern",
    "NodeFreeEvent",
    "NoteEvent",
    "NullEvent",
    "ParallelPattern",
    "Pattern",
    "PatternPlayer",
    "RandomPattern",
    "RepeatPattern",
    "RestartPattern",
    "SeedPattern",
    "SequencePattern",
    "ShufflePattern",
    "StutterPattern",
    "SynthAllocateEvent",
    "UnaryOpPattern",
    "UpdatePattern",
]



import collections
import copy
import inspect
import threading
import uuid
from collections.abc import Sequence
from typing import List, Tuple, Union

from uqbar.objects import new

from supriya.enums import ParameterRate
from supriya.system import SupriyaObject

from .controls import Parameter

_local = threading.local()
_local._active_builders = []


class SynthDefBuilder(SupriyaObject):
    """
    A SynthDef builder.

    ::

        >>> import supriya.synthdefs
        >>> import supriya.ugens

    ::

        >>> builder = supriya.synthdefs.SynthDefBuilder(
        ...     frequency=440,
        ...     trigger=supriya.synthdefs.Parameter(
        ...         value=0,
        ...         parameter_rate=supriya.ParameterRate.TRIGGER,
        ...     ),
        ... )

    ::

        >>> with builder:
        ...     sin_osc = supriya.ugens.SinOsc.ar(
        ...         frequency=builder["frequency"],
        ...     )
        ...     decay = supriya.ugens.Decay.kr(
        ...         decay_time=0.5,
        ...         source=builder["trigger"],
        ...     )
        ...     enveloped_sin = sin_osc * decay
        ...     out = supriya.ugens.Out.ar(bus=0, source=enveloped_sin)
        ...

    ::

        >>> synthdef = builder.build()
        >>> supriya.graph(synthdef)  # doctest: +SKIP

    """

    ### CLASS VARIABLES ###

    _active_builders: List["SynthDefBuilder"] = _local._active_builders

    __slots__ = ("_name", "_parameters", "_ugens", "_uuid")

    ### INITIALIZER ###

    def __init__(self, name=None, **kwargs):
        self._name = name
        self._uuid = uuid.uuid4()
        self._parameters = collections.OrderedDict()
        self._ugens = []
        for key, value in kwargs.items():
            self._add_parameter(key, value)

    ### SPECIAL METHODS ###

    def __enter__(self):
        SynthDefBuilder._active_builders.append(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        SynthDefBuilder._active_builders.pop()

    def __getitem__(self, item):
        return self._parameters[item]

    ### PRIVATE METHODS ###

    def _add_ugens(self, ugens):
        import supriya.synthdefs
        import supriya.ugens

        if not isinstance(ugens, Sequence):
            ugens = [ugens]
        prototype = (
            supriya.synthdefs.OutputProxy,
            supriya.synthdefs.Parameter,
            supriya.synthdefs.UGen,
        )
        for ugen in ugens:
            assert isinstance(ugen, prototype), type(ugen)
            if isinstance(ugen, supriya.synthdefs.OutputProxy):
                ugen = ugen.source
            assert ugen._uuid == self._uuid
            if ugen not in self._ugens:
                self._ugens.append(ugen)

    def _add_parameter(self, *args):
        # TODO: Refactor without *args for clarity
        import supriya.synthdefs

        if 3 < len(args):
            raise ValueError(args)
        if len(args) == 1:
            assert isinstance(args[0], supriya.synthdefs.Parameter)
            name, value, parameter_rate = args[0].name, args[0], args[0].parameter_rate
        elif len(args) == 2:
            name, value = args
            if isinstance(value, supriya.synthdefs.Parameter):
                parameter_rate = value.parameter_rate
            else:
                parameter_rate = supriya.ParameterRate.CONTROL
                if name.startswith("a_"):
                    parameter_rate = supriya.ParameterRate.AUDIO
                elif name.startswith("i_"):
                    parameter_rate = supriya.ParameterRate.SCALAR
                elif name.startswith("t_"):
                    parameter_rate = supriya.ParameterRate.TRIGGER
        elif len(args) == 3:
            name, value, parameter_rate = args
            parameter_rate = supriya.ParameterRate.from_expr(parameter_rate)
        else:
            raise ValueError(args)
        if not isinstance(value, supriya.synthdefs.Parameter):
            parameter = supriya.synthdefs.Parameter(
                name=name, parameter_rate=parameter_rate, value=value
            )
        else:
            parameter = new(value, parameter_rate=parameter_rate, name=name)
        assert parameter._uuid is None
        parameter._uuid = self._uuid
        self._parameters[name] = parameter
        return parameter

    ### PUBLIC METHODS ###

    def build(self, name=None, optimize=True):
        import supriya.synthdefs
        import supriya.ugens

        # Calling build() creates controls each time, so strip out
        # previously created ones. This could be made cleaner by preventing
        # Control subclasses from being aggregated into SynthDefBuilders in
        # the first place.

        self._ugens[:] = [
            ugen
            for ugen in self._ugens
            if not isinstance(ugen, supriya.synthdefs.Control)
        ]
        name = self.name or name
        with self:
            ugens = list(self._parameters.values()) + list(self._ugens)
            ugens = copy.deepcopy(ugens)
            ugens, parameters = supriya.synthdefs.SynthDef._extract_parameters(ugens)
            (
                control_ugens,
                control_mapping,
                indexed_parameters,
            ) = supriya.synthdefs.SynthDef._build_control_mapping(parameters)
            supriya.synthdefs.SynthDef._remap_controls(ugens, control_mapping)
            ugens = control_ugens + ugens
            synthdef = supriya.synthdefs.SynthDef(ugens, name=name, optimize=optimize)
        return synthdef

    def poll_ugen(self, ugen, label=None, trigger=None, trigger_id=-1):
        import supriya.ugens

        if trigger is None:
            trigger = supriya.ugens.Impulse.kr(1)
        poll = supriya.ugens.Poll.new(
            source=ugen, label=label, trigger=trigger, trigger_id=trigger_id
        )
        self._add_ugens(poll)

    ### PUBLIC PROPERTIES ###

    @property
    def name(self):
        return self._name


def synthdef(*args: Union[str, Tuple[str, float]]):
    """
    Decorate for quickly constructing SynthDefs from functions.

    ::

        >>> from supriya.ugens import EnvGen, Out, SinOsc
        >>> from supriya.synthdefs import Envelope, synthdef

    ::

        >>> @synthdef()
        ... def sine(freq=440, amp=0.1, gate=1):
        ...     sig = SinOsc.ar(freq) * amp
        ...     env = EnvGen.kr(envelope=Envelope.adsr(), gate=gate, done_action=2)
        ...     Out.ar(0, [sig * env] * 2)
        ...

    ::

        >>> print(sine)
        synthdef:
            name: sine
            ugens:
            -   Control.kr: null
            -   SinOsc.ar:
                    frequency: Control.kr[1:freq]
                    phase: 0.0
            -   BinaryOpUGen(MULTIPLICATION).ar/0:
                    left: SinOsc.ar[0]
                    right: Control.kr[0:amp]
            -   EnvGen.kr:
                    done_action: 2.0
                    envelope[0]: 0.0
                    envelope[10]: 5.0
                    envelope[11]: -4.0
                    envelope[12]: 0.0
                    envelope[13]: 1.0
                    envelope[14]: 5.0
                    envelope[15]: -4.0
                    envelope[1]: 3.0
                    envelope[2]: 2.0
                    envelope[3]: -99.0
                    envelope[4]: 1.0
                    envelope[5]: 0.01
                    envelope[6]: 5.0
                    envelope[7]: -4.0
                    envelope[8]: 0.5
                    envelope[9]: 0.3
                    gate: Control.kr[2:gate]
                    level_bias: 0.0
                    level_scale: 1.0
                    time_scale: 1.0
            -   BinaryOpUGen(MULTIPLICATION).ar/1:
                    left: BinaryOpUGen(MULTIPLICATION).ar/0[0]
                    right: EnvGen.kr[0]
            -   Out.ar:
                    bus: 0.0
                    source[0]: BinaryOpUGen(MULTIPLICATION).ar/1[0]
                    source[1]: BinaryOpUGen(MULTIPLICATION).ar/1[0]

    ::

        >>> @synthdef("ar", ("kr", 0.5))
        ... def sine(freq=440, amp=0.1, gate=1):
        ...     sig = SinOsc.ar(freq) * amp
        ...     env = EnvGen.kr(envelope=Envelope.adsr(), gate=gate, done_action=2)
        ...     Out.ar(0, [sig * env] * 2)
        ...

    ::

        >>> print(sine)
        synthdef:
            name: sine
            ugens:
            -   AudioControl.ar: null
            -   SinOsc.ar:
                    frequency: AudioControl.ar[0:freq]
                    phase: 0.0
            -   LagControl.kr:
                    lags[0]: 0.5
                    lags[1]: 0.0
            -   BinaryOpUGen(MULTIPLICATION).ar/0:
                    left: SinOsc.ar[0]
                    right: LagControl.kr[0:amp]
            -   EnvGen.kr:
                    done_action: 2.0
                    envelope[0]: 0.0
                    envelope[10]: 5.0
                    envelope[11]: -4.0
                    envelope[12]: 0.0
                    envelope[13]: 1.0
                    envelope[14]: 5.0
                    envelope[15]: -4.0
                    envelope[1]: 3.0
                    envelope[2]: 2.0
                    envelope[3]: -99.0
                    envelope[4]: 1.0
                    envelope[5]: 0.01
                    envelope[6]: 5.0
                    envelope[7]: -4.0
                    envelope[8]: 0.5
                    envelope[9]: 0.3
                    gate: LagControl.kr[1:gate]
                    level_bias: 0.0
                    level_scale: 1.0
                    time_scale: 1.0
            -   BinaryOpUGen(MULTIPLICATION).ar/1:
                    left: BinaryOpUGen(MULTIPLICATION).ar/0[0]
                    right: EnvGen.kr[0]
            -   Out.ar:
                    bus: 0.0
                    source[0]: BinaryOpUGen(MULTIPLICATION).ar/1[0]
                    source[1]: BinaryOpUGen(MULTIPLICATION).ar/1[0]

    """

    def inner(func):
        signature = inspect.signature(func)
        builder = SynthDefBuilder()
        kwargs = {}
        for i, (name, parameter) in enumerate(signature.parameters.items()):
            rate = ParameterRate.CONTROL
            lag = None
            try:
                if isinstance(args[i], str):
                    rate_expr = args[i]
                else:
                    rate_expr, lag = args[i]
                rate = ParameterRate.from_expr(rate_expr)
            except (IndexError, TypeError):
                pass
            value = parameter.default
            if value is inspect._empty:
                value = 0.0
            parameter = Parameter(lag=lag, name=name, parameter_rate=rate, value=value)
            kwargs[name] = builder._add_parameter(parameter)
        with builder:
            func(**kwargs)
        return builder.build(name=func.__name__)

    return inner



import asyncio
import atexit
import logging
import os
import signal
import subprocess
import time

import supriya.exceptions

logger = logging.getLogger("supriya.server.protocol")


class ProcessProtocol:
    def __init__(self):
        self.is_running = False
        atexit.register(self.quit)

    def boot(self, options, scsynth_path, port):
        ...

    def quit(self):
        ...


class SyncProcessProtocol(ProcessProtocol):

    ### PUBLIC METHODS ###

    def boot(self, options, scsynth_path, port):
        if self.is_running:
            return
        options_string = options.as_options_string(port)
        command = "{} {}".format(scsynth_path, options_string)
        logger.info("Boot: {}".format(command))
        try:
            self.process = subprocess.Popen(
                command,
                shell=True,
                stderr=subprocess.STDOUT,
                stdout=subprocess.PIPE,
                start_new_session=True,
            )
            start_time = time.time()
            timeout = 10
            while True:
                line = self.process.stdout.readline().decode().rstrip()
                if line:
                    logger.info("Boot: {}".format(line))
                if line.startswith("SuperCollider 3 server ready"):
                    break
                elif line.startswith("Supernova ready"):
                    break
                elif line.startswith(("ERROR:", "*** ERROR:")):
                    raise supriya.exceptions.ServerCannotBoot(line)
                elif line.startswith(
                    "Exception in World_OpenUDP: bind: Address already in use"
                ):
                    raise supriya.exceptions.ServerCannotBoot(line)
                elif (time.time() - start_time) > timeout:
                    raise supriya.exceptions.ServerCannotBoot(line)
            self.is_running = True
        except supriya.exceptions.ServerCannotBoot:
            try:
                process_group = os.getpgid(self.process.pid)
                os.killpg(process_group, signal.SIGINT)
                self.process.terminate()
                self.process.wait()
            except ProcessLookupError:
                pass
            raise

    def quit(self):
        if not self.is_running:
            return
        # try:
        #    self.process.communicate(timeout=0.1)
        # except subprocess.TimeoutExpired:
        #    self.process.kill()
        #    self.process.communicate()
        try:
            process_group = os.getpgid(self.process.pid)
            os.killpg(process_group, signal.SIGINT)
        except ProcessLookupError:
            logger.warning(f"Could not find process group for PID {self.process.pid}")
        self.process.terminate()
        self.process.wait()
        self.is_running = False


class AsyncProcessProtocol(asyncio.SubprocessProtocol, ProcessProtocol):

    ### INITIALIZER ###

    def __init__(self):
        ProcessProtocol.__init__(self)
        asyncio.SubprocessProtocol.__init__(self)
        self.boot_future = None
        self.exit_future = None

    ### PUBLIC METHODS ###

    async def boot(self, options, scsynth_path, port):
        if self.is_running:
            return
        self.is_running = False
        options_string = options.as_options_string(port)
        command = "{} {}".format(scsynth_path, options_string)
        logger.info(command)
        loop = asyncio.get_running_loop()
        self.boot_future = loop.create_future()
        self.exit_future = loop.create_future()
        _, _ = await loop.subprocess_exec(
            lambda: self, *command.split(), stdin=None, stderr=None
        )

    def connection_made(self, transport):
        self.is_running = True
        self.transport = transport

    def pipe_data_received(self, fd, data):
        for line in data.splitlines():
            logger.info(line.decode())
            if line.strip().startswith(b"Exception"):
                self.boot_future.set_result(False)
            elif line.strip().startswith(b"SuperCollider 3 server ready"):
                self.boot_future.set_result(True)

    def process_exited(self):
        self.is_running = False
        self.exit_future.set_result(None)
        if not self.boot_future.done():
            self.boot_future.set_result(False)

    def quit(self):
        if not self.is_running:
            return
        if not self.boot_future.done():
            self.boot_future.set_result(False)
        if not self.exit_future.done():
            self.exit_future.set_result
        if not self.transport._loop.is_closed() and not self.transport.is_closing():
            self.transport.close()
        self.is_running = False



import supriya.system
from supriya import utils
from supriya.system import SupriyaObject


class Meters(SupriyaObject):
    """
    Server meters.

    ::

        >>> import json, supriya, time
        >>> server = supriya.Server().boot(
        ...     input_bus_channel_count=2,
        ...     output_bus_channel_count=2,
        ... )
        >>> meters = server.meters.allocate()
        >>> time.sleep(1)
        >>> print(json.dumps(meters.to_dict(), indent=4))
        {
            "server_meters": {
                "input_meter_levels": [
                    {
                        "peak": ...,
                        "rms": ...
                    },
                    {
                        "peak": ...,
                        "rms": ...
                    }
                ],
                "output_meter_levels": [
                    {
                        "peak": 0.0,
                        "rms": 0.0
                    },
                    {
                        "peak": 0.0,
                        "rms": 0.0
                    }
                ]
            }
        }

    ::

        >>> meters.free()

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_input_meter_callback",
        "_input_meter_peak_levels",
        "_input_meter_rms_levels",
        "_input_meter_synth",
        "_output_meter_callback",
        "_output_meter_peak_levels",
        "_output_meter_rms_levels",
        "_output_meter_synth",
        "_server",
    )

    ### INITIALIZER ###

    def __init__(self, server):
        self._server = server
        self._input_meter_callback = None
        self._input_meter_peak_levels = None
        self._input_meter_rms_levels = None
        self._input_meter_synth = None
        self._output_meter_callback = None
        self._output_meter_peak_levels = None
        self._output_meter_rms_levels = None
        self._output_meter_synth = None

    ### PRIVATE METHODS ###

    def _handle_input_levels(self, message):
        contents = message.contents[2:]
        peak_levels = []
        rms_levels = []
        for peak, rms in utils.group_iterable_by_count(contents, 2):
            peak_levels.append(peak)
            rms_levels.append(rms)
        self._input_meter_peak_levels = tuple(peak_levels)
        self._input_meter_rms_levels = tuple(rms_levels)

    def _handle_output_levels(self, message):
        contents = message.contents[2:]
        peak_levels = []
        rms_levels = []
        for peak, rms in utils.group_iterable_by_count(contents, 2):
            peak_levels.append(peak)
            rms_levels.append(rms)
        self._output_meter_peak_levels = tuple(peak_levels)
        self._output_meter_rms_levels = tuple(rms_levels)

    ### PUBLIC METHODS ###

    def allocate(self):
        import supriya.osc
        import supriya.realtime

        self._input_meter_callback = self.server.osc_protocol.register(
            pattern=self.input_meter_command, procedure=self._handle_input_levels
        )
        self._output_meter_callback = self.server.osc_protocol.register(
            pattern=self.output_meter_command, procedure=self._handle_output_levels
        )
        input_meter_synthdef = self.input_meter_synthdef
        output_meter_synthdef = self.output_meter_synthdef
        self._input_meter_synth = supriya.realtime.Synth(input_meter_synthdef)
        self._output_meter_synth = supriya.realtime.Synth(output_meter_synthdef)
        self._input_meter_synth.allocate(
            add_action=supriya.AddAction.ADD_TO_HEAD,
            node_id_is_permanent=True,
            target_node=self.server.root_node,
        )
        self._output_meter_synth.allocate(
            add_action=supriya.AddAction.ADD_TO_TAIL,
            node_id_is_permanent=True,
            target_node=self.server.root_node,
        )
        return self

    def free(self):
        self.server.osc_protocol.unregister(self._input_meter_callback)
        self.server.osc_protocol.unregister(self._output_meter_callback)
        self._input_meter_synth.free()
        self._output_meter_synth.free()
        self._input_meter_callback = None
        self._output_meter_callback = None
        self._input_meter_synth = None
        self._output_meter_synth = None

    @staticmethod
    def make_meter_synthdef(channel_count=1, command_name="/reply", initial_bus=0):
        import supriya.synthdefs
        import supriya.ugens

        with supriya.synthdefs.SynthDefBuilder() as builder:
            source = supriya.ugens.In.ar(bus=initial_bus, channel_count=channel_count)
            supriya.ugens.SendPeakRMS.kr(
                command_name=command_name, peak_lag=1, reply_rate=20, source=source
            )
        synthdef = builder.build()
        return synthdef

    def notify(self, topic, event):
        if topic == "server-quitting":
            self.free()

    def to_dict(self):
        if not self.is_allocated:
            raise supriya.exceptions.NotAllocated(self)
        input_meter_levels, output_meter_levels = [], []
        for peak, rms in zip(
            self._input_meter_peak_levels or [], self._input_meter_rms_levels or []
        ):
            input_meter_levels.append(dict(peak=peak, rms=rms))
        for peak, rms in zip(
            self._output_meter_peak_levels or [], self._output_meter_rms_levels or []
        ):
            output_meter_levels.append(dict(peak=peak, rms=rms))
        result = {
            "server_meters": {
                "input_meter_levels": input_meter_levels,
                "output_meter_levels": output_meter_levels,
            }
        }
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def input_count(self):
        return self.server.options.input_bus_channel_count

    @property
    def is_allocated(self):
        return self.server is not None and self._input_meter_synth in self.server

    @property
    def output_count(self):
        return self.server.options.output_bus_channel_count

    @property
    def input_meter_command(self):
        return "/meter.inputs"

    @property
    def input_meter_synthdef(self):
        return self.make_meter_synthdef(
            channel_count=self.server.options.input_bus_channel_count,
            initial_bus=self.server.options.output_bus_channel_count,
            command_name=self.input_meter_command,
        )

    @property
    def output_meter_command(self):
        return "/meter.outputs"

    @property
    def output_meter_synthdef(self):
        return self.make_meter_synthdef(
            channel_count=self.server.options.output_bus_channel_count,
            initial_bus=0,
            command_name=self.output_meter_command,
        )

    @property
    def server(self):
        return self._server



import collections

from supriya.enums import CalculationRate
from supriya.synthdefs import MultiOutUGen, UGen


class FreqShift(UGen):
    """

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> freq_shift = supriya.ugens.FreqShift.ar(
        ...     frequency=0,
        ...     phase=0,
        ...     source=source,
        ... )
        >>> freq_shift
        FreqShift.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("frequency", 0.0), ("phase", 0.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class Hilbert(MultiOutUGen):
    """
    Applies the Hilbert transform.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> hilbert = supriya.ugens.Hilbert.ar(
        ...     source=source,
        ... )
        >>> hilbert
        UGenArray({2})

    """

    _default_channel_count = 2
    _has_settable_channel_count = False
    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class HilbertFIR(UGen):
    """
    Applies the Hilbert transform.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> hilbert_fir = supriya.ugens.HilbertFIR.ar(
        ...     buffer_id=23,
        ...     source=source,
        ... )
        >>> hilbert_fir
        HilbertFIR.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("buffer_id", None)]
    )

    _valid_calculation_rates = (CalculationRate.AUDIO,)



import collections

from supriya import CalculationRate
from supriya.synthdefs import UGen


class Done(UGen):
    """
    Triggers when `source` sets its `done` flag.

    ::

        >>> source = supriya.ugens.Line.kr()
        >>> done = supriya.ugens.Done.kr(
        ...     source=source,
        ... )
        >>> done
        Done.kr()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.CONTROL,)

    ### INITIALIZER ###

    def __init__(self, calculation_rate=None, source=None):
        if not (hasattr(source, "has_done_flag") and source.has_done_flag):
            raise ValueError(repr(source))
        UGen.__init__(self, calculation_rate=calculation_rate, source=source)


class EnvGen(UGen):
    """
    An envelope generator.

    ::

        >>> envelope = supriya.synthdefs.Envelope.percussive()
        >>> supriya.ugens.EnvGen.ar(envelope=envelope)
        EnvGen.ar()

    """

    ### CLASS VARIABLES ###

    _has_done_flag = True
    _ordered_input_names = collections.OrderedDict(
        [
            ("gate", 1.0),
            ("level_scale", 1.0),
            ("level_bias", 0.0),
            ("time_scale", 1.0),
            ("done_action", 0.0),
            ("envelope", None),
        ]
    )
    _unexpanded_input_names = ("envelope",)
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)

    ### PRIVATE METHODS ###

    @classmethod
    def _new_expanded(
        cls,
        calculation_rate=None,
        done_action=None,
        envelope=None,
        gate=1.0,
        level_bias=0.0,
        level_scale=1.0,
        time_scale=1.0,
    ):
        import supriya.synthdefs

        if not isinstance(done_action, supriya.synthdefs.Parameter):
            done_action = supriya.DoneAction.from_expr(done_action)
        if envelope is None:
            envelope = supriya.synthdefs.Envelope()
        assert isinstance(envelope, supriya.synthdefs.Envelope)
        envelope = envelope.serialize()
        return super(EnvGen, cls)._new_expanded(
            calculation_rate=calculation_rate,
            done_action=done_action,
            envelope=envelope,
            gate=gate,
            level_bias=level_bias,
            level_scale=level_scale,
            time_scale=time_scale,
        )


class Free(UGen):
    """
    Frees the node at `node_id` when triggered by `trigger`.

    ::

        >>> node_id = 1000
        >>> trigger = supriya.ugens.Impulse.kr(frequency=1.0)
        >>> free = supriya.ugens.Free.kr(
        ...     node_id=node_id,
        ...     trigger=trigger,
        ... )
        >>> free
        Free.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("trigger", 0.0), ("node_id", None)]
    )
    _valid_calculation_rates = (CalculationRate.CONTROL,)


class FreeSelf(UGen):
    """
    Frees the enclosing synth when triggered by `trigger`.

    ::

        >>> trigger = supriya.ugens.Impulse.kr(frequency=1.0)
        >>> free_self = supriya.ugens.FreeSelf.kr(
        ...     trigger=trigger,
        ... )
        >>> free_self
        FreeSelf.kr()

    """

    _ordered_input_names = collections.OrderedDict([("trigger", None)])
    _valid_calculation_rates = (CalculationRate.CONTROL,)


class FreeSelfWhenDone(UGen):
    """
    Frees the enclosing synth when `source` sets its `done` flag.

    ::

        >>> source = supriya.ugens.Line.kr()
        >>> free_self_when_done = supriya.ugens.FreeSelfWhenDone.kr(
        ...     source=source,
        ... )
        >>> free_self_when_done
        FreeSelfWhenDone.kr()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.CONTROL,)

    ### INITIALIZER ###

    def __init__(self, calculation_rate=None, source=None):
        if not (hasattr(source, "has_done_flag") and source.has_done_flag):
            raise ValueError(repr(source))
        UGen.__init__(self, calculation_rate=calculation_rate, source=source)


class Pause(UGen):
    """
    Pauses the node at `node_id` when triggered by `trigger`.

    ::

        >>> node_id = 1000
        >>> trigger = supriya.ugens.Impulse.kr(frequency=1.0)
        >>> pause = supriya.ugens.Pause.kr(
        ...     node_id=node_id,
        ...     trigger=trigger,
        ... )
        >>> pause
        Pause.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("trigger", None), ("node_id", None)]
    )
    _valid_calculation_rates = (CalculationRate.CONTROL,)


class PauseSelf(UGen):
    """
    Pauses the enclosing synth when triggered by `trigger`.

    ::

        >>> trigger = supriya.ugens.Impulse.kr(frequency=1.0)
        >>> pause_self = supriya.ugens.PauseSelf.kr(
        ...     trigger=trigger,
        ... )
        >>> pause_self
        PauseSelf.kr()

    """

    _ordered_input_names = collections.OrderedDict([("trigger", None)])
    _valid_calculation_rates = (CalculationRate.CONTROL,)


class PauseSelfWhenDone(UGen):
    """
    Pauses the enclosing synth when `source` sets its `done` flag.

    ::

        >>> source = supriya.ugens.Line.kr()
        >>> pause_self_when_done = supriya.ugens.PauseSelfWhenDone.kr(
        ...     source=source,
        ... )
        >>> pause_self_when_done
        PauseSelfWhenDone.kr()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.CONTROL,)

    ### INITIALIZER ###

    def __init__(self, calculation_rate=None, source=None):
        if not (hasattr(source, "has_done_flag") and source.has_done_flag):
            raise ValueError(repr(source))
        UGen.__init__(self, calculation_rate=calculation_rate, source=source)


class Linen(UGen):
    """
    A simple line generating unit generator.

    ::

        >>> supriya.ugens.Linen.kr()
        Linen.kr()

    """

    _has_done_flag = True
    _ordered_input_names = collections.OrderedDict(
        [
            ("gate", 1.0),
            ("attack_time", 0.01),
            ("sustain_level", 1.0),
            ("release_time", 1.0),
            ("done_action", 0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.CONTROL,)



from IPython.core.display import display  # type: ignore
from IPython.display import Audio  # type: ignore

from supriya.ext import websafe_audio
from supriya.io import Player


def load_ipython_extension(ipython):
    ipython.extension_manager.load_extension("uqbar.ext.ipython")
    patch_player()


def patch_player():
    def render(self):
        output_path = self.renderable.__render__(**self.render_kwargs)
        return websafe_audio(output_path)

    def open_output_path(self, output_path):
        display(Audio(filename=str(output_path)))

    Player.open_output_path = open_output_path
    Player.render = render



import pathlib

import supriya.osc
from supriya.enums import RequestId

from .bases import Request, RequestBundle, Response


class SynthDefFreeAllRequest(Request):

    ### CLASS VARIABLES ###

    request_id = RequestId.SYNTHDEF_FREE_ALL

    ### INITIALIZER ###

    def __init__(self):
        Request.__init__(self)
        raise NotImplementedError

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        raise NotImplementedError


class SynthDefFreeRequest(Request):
    """
    A /d_free request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.SynthDefFreeRequest(
        ...     synthdef="test",
        ... )
        >>> request
        SynthDefFreeRequest(
            synthdef='test',
        )

    ::

        >>> request.to_osc()
        OscMessage('/d_free', 'test')

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.SYNTHDEF_FREE

    ### INITIALIZER ###

    def __init__(self, synthdef=None):
        import supriya.synthdefs

        Request.__init__(self)
        prototype = (str, supriya.synthdefs.SynthDef)
        assert isinstance(synthdef, prototype)
        self._synthdef = synthdef

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        import supriya.synthdefs

        request_id = self.request_name
        synthdef = self.synthdef
        if isinstance(synthdef, supriya.synthdefs.SynthDef):
            synthdef = synthdef.actual_name
        message = supriya.osc.OscMessage(request_id, synthdef)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def synthdef(self):
        return self._synthdef


class SynthDefLoadDirectoryRequest(Request):
    """
    A /d_loadDir request.
    """

    ### CLASS VARIABLES ###

    request_id = RequestId.SYNTHDEF_LOAD_DIR

    ### INITIALIZER ###

    def __init__(self, callback=None, directory_path=None):
        Request.__init__(self)
        if callback is not None:
            assert isinstance(callback, (Request, RequestBundle))
        self._callback = callback
        self._directory_path = pathlib.Path(directory_path).absolute()

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        contents = [request_id, str(self.directory_path)]
        if self.callback:
            contents.append(self.callback.to_osc(with_placeholders=with_placeholders))
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def callback(self):
        return self._callback

    @property
    def response_patterns(self):
        return ["/done", "/d_loadDir"], None

    @property
    def directory_path(self):
        return self._directory_path


class SynthDefLoadRequest(Request):
    """
    A /d_load request.
    """

    ### CLASS VARIABLES ###

    request_id = RequestId.SYNTHDEF_LOAD

    ### INITIALIZER ###

    def __init__(self, callback=None, synthdef_path=None):
        Request.__init__(self)
        if callback is not None:
            assert isinstance(callback, (Request, RequestBundle))
        self._callback = callback
        self._synthdef_path = pathlib.Path(synthdef_path).absolute()

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        contents = [request_id, str(self.synthdef_path)]
        if self.callback:
            contents.append(self.callback.to_osc(with_placeholders=with_placeholders))
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def callback(self):
        return self._callback

    @property
    def response_patterns(self):
        return ["/done", "/d_load"], None

    @property
    def synthdef_path(self):
        return self._synthdef_path


class SynthDefReceiveRequest(Request):
    """
    A /d_recv request.

    ::

        >>> server = supriya.Server().boot()


    ::

        >>> with supriya.SynthDefBuilder(out=0, value=0.5) as builder:
        ...     _ = supriya.ugens.Out.ar(
        ...         bus=builder["out"],
        ...         source=supriya.ugens.DC.ar(builder["value"]),
        ...     )
        ...
        >>> synthdef = builder.build(name="example")

    ::

        >>> synthdef in server
        False

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group

    Allocate a synthdef, then allocate a new group and allocate a synth in that
    group using the newly allocated synthdef:

    ::

        >>> request = supriya.commands.SynthDefReceiveRequest(
        ...     synthdefs=[synthdef],
        ...     callback=supriya.commands.RequestBundle(
        ...         contents=[
        ...             supriya.commands.GroupNewRequest(
        ...                 items=[
        ...                     supriya.commands.GroupNewRequest.Item(
        ...                         node_id=1000,
        ...                         target_node_id=1,
        ...                     ),
        ...                 ],
        ...             ),
        ...             supriya.commands.SynthNewRequest(
        ...                 node_id=1001,
        ...                 synthdef=synthdef,
        ...                 target_node_id=1000,
        ...             ),
        ...         ],
        ...     ),
        ... )

    ::

        >>> with server.osc_protocol.capture() as transcript:
        ...     response = request.communicate(server=server)
        ...     _ = server.sync()
        ...
        >>> response
        DoneResponse(
            action=('/d_recv',),
        )

    ::

        >>> for entry in transcript:
        ...     (entry.label, entry.message)
        ...
        ('S', OscMessage('/d_recv', bytearray(b'SCgf...example...'), OscBundle(
            contents=(
                OscMessage('/g_new', 1000, 0, 1),
                OscMessage('/s_new', 'example', 1001, 0, 1000),
            ),
        )))
        ('R', OscMessage('/n_go', 1000, 1, -1, -1, 1, -1, -1))
        ('R', OscMessage('/n_go', 1001, 1000, -1, -1, 0))
        ('R', OscMessage('/done', '/d_recv'))
        ('S', OscMessage('/sync', 0))
        ('R', OscMessage('/synced', 0))

    ::

        >>> print(server.query())
        NODE TREE 0 group
            1 group
                1000 group
                    1001 example
                        out: 0.0, value: 0.5

    ::

        >>> print(server.root_node)
        NODE TREE 0 group
            1 group
                1000 group
                    1001 example
                        out: 0.0, value: 0.5

    ::

        >>> synthdef in server
        True

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.SYNTHDEF_RECEIVE

    ### INITIALIZER ###

    def __init__(self, callback=None, synthdefs=None, use_anonymous_names=None):
        import supriya.synthdefs

        Request.__init__(self)
        if callback is not None:
            assert isinstance(callback, (Request, RequestBundle))
        self._callback = callback
        if synthdefs:
            prototype = supriya.synthdefs.SynthDef
            if isinstance(synthdefs, prototype):
                synthdefs = (synthdefs,)
            assert all(isinstance(x, prototype) for x in synthdefs)
            synthdefs = tuple(synthdefs)
        self._synthdefs = synthdefs
        if use_anonymous_names is not None:
            use_anonymous_names = bool(use_anonymous_names)
        self._use_anonymous_names = use_anonymous_names

    ### PRIVATE METHODS ###

    def _apply_local(self, server):
        for synthdef in self.synthdefs:
            synthdef._register_with_local_server(server)

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        import supriya.synthdefs

        request_id = self.request_name
        compiled_synthdefs = supriya.synthdefs.SynthDefCompiler.compile_synthdefs(
            self.synthdefs, use_anonymous_names=self.use_anonymous_names
        )
        compiled_synthdefs = bytearray(compiled_synthdefs)
        contents = [request_id, compiled_synthdefs]
        if self.callback:
            contents.append(self.callback.to_osc(with_placeholders=with_placeholders))
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def callback(self):
        return self._callback

    @property
    def response_patterns(self):
        return ["/done", "/d_recv"], None

    @property
    def synthdefs(self):
        return self._synthdefs

    @property
    def use_anonymous_names(self):
        return self._use_anonymous_names


class SynthDefRemovedResponse(Response):

    ### INITIALIZER ###

    def __init__(self, synthdef_name=None):
        Response.__init__(self)
        self._synthdef_name = synthdef_name

    ### PUBLIC METHODS ###

    @classmethod
    def from_osc_message(cls, osc_message):
        synthdef_name = osc_message.contents[0]
        response = cls(synthdef_name=synthdef_name)
        return response

    ### PUBLIC PROPERTIES ###

    @property
    def synthdef_name(self):
        return self._synthdef_name



import collections

from supriya import CalculationRate
from supriya.synthdefs import UGen


class Convolution(UGen):
    """
    A real-time convolver.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> kernel = supriya.ugens.Mix.new(
        ...     supriya.ugens.LFSaw.ar(frequency=[300, 500, 800, 1000])
        ...     * supriya.ugens.MouseX.kr(minimum=1, maximum=2),
        ... )
        >>> convolution = supriya.ugens.Convolution.ar(
        ...     framesize=512,
        ...     kernel=kernel,
        ...     source=source,
        ... )
        >>> convolution
        Convolution.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("kernel", None), ("framesize", 512)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class Convolution2(UGen):
    """
    Strict convolution with fixed kernel which can be updated using a trigger
    signal.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> kernel = supriya.ugens.Mix.new(
        ...     supriya.ugens.LFSaw.ar(frequency=[300, 500, 800, 1000])
        ...     * supriya.ugens.MouseX.kr(minimum=1, maximum=2),
        ... )
        >>> convolution_2 = supriya.ugens.Convolution2.ar(
        ...     framesize=2048,
        ...     kernel=kernel,
        ...     source=source,
        ...     trigger=0,
        ... )
        >>> convolution_2
        Convolution2.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("kernel", None), ("trigger", 0), ("framesize", 2048)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class Convolution2L(UGen):
    """
    Strict convolution with fixed kernel which can be updated using a trigger signal.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> kernel = supriya.ugens.Mix.new(
        ...     supriya.ugens.LFSaw.ar(frequency=[300, 500, 800, 1000])
        ...     * supriya.ugens.MouseX.kr(minimum=1, maximum=2),
        ... )
        >>> convolution_2_l = supriya.ugens.Convolution2L.ar(
        ...     crossfade=1,
        ...     framesize=2048,
        ...     kernel=kernel,
        ...     source=source,
        ...     trigger=0,
        ... )
        >>> convolution_2_l
        Convolution2L.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("source", None),
            ("kernel", None),
            ("trigger", 0.0),
            ("framesize", 2048),
            ("crossfade", 1.0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class Convolution3(UGen):
    """
    Strict convolution with fixed kernel which can be updated using a trigger signal.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> kernel = supriya.ugens.Mix.new(
        ...     supriya.ugens.LFSaw.ar(frequency=[300, 500, 800, 1000])
        ...     * supriya.ugens.MouseX.kr(minimum=1, maximum=2),
        ... )
        >>> convolution_3 = supriya.ugens.Convolution3.ar(
        ...     framesize=2048,
        ...     kernel=kernel,
        ...     source=source,
        ...     trigger=0,
        ... )
        >>> convolution_3
        Convolution3.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("kernel", None), ("trigger", 0.0), ("framesize", 2048)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)



from uqbar.enums import IntEnumeration

from .patterns import Pattern, SequencePattern


class ChoicePattern(SequencePattern):
    def __init__(self, sequence, iterations=1, forbid_repetitions=False, weights=None):
        super().__init__(sequence, iterations=iterations)
        self._forbid_repetitions = bool(forbid_repetitions)
        if weights:
            weights = tuple(abs(float(x)) for x in weights)
        self._weights = weights or None

    def _iterate(self, state=None):
        should_stop = False
        rng = self._get_rng()
        previous_index = None
        for _ in self._loop(self._iterations):
            if self.weights:
                index = self._find_index_weighted(rng)
                while self.forbid_repetitions and index == previous_index:
                    index = self._find_index_weighted(rng)
            else:
                index = self._find_index_unweighted(rng)
                while self.forbid_repetitions and index == previous_index:
                    index = self._find_index_unweighted(rng)
            previous_index = index
            choice = self._sequence[index]
            if isinstance(choice, Pattern):
                should_stop = (yield from choice) or should_stop
            else:
                should_stop = (yield choice) or should_stop
            if should_stop:
                return

    def _find_index_unweighted(self, rng):
        return int(next(rng) * 0x7FFFFFFF) % len(self._sequence)

    def _find_index_weighted(self, rng):
        needle = next(rng) * sum(self.weights)
        sum_of_weights = 0.0
        for index, weight in enumerate(self.weights):
            if sum_of_weights <= needle <= sum_of_weights + weight:
                break
            sum_of_weights += weight
        return index

    @property
    def forbid_repetitions(self):
        return self._forbid_repetitions

    @property
    def weights(self):
        return self._weights


class RandomPattern(Pattern):
    class Distribution(IntEnumeration):
        WHITE_NOISE = 0

    def __init__(
        self, minimum=0.0, maximum=1.0, iterations=None, distribution="WHITE_NOISE"
    ):
        if iterations is not None:
            iterations = int(iterations)
            if iterations < 1:
                raise ValueError("Iterations must be null or greater than 0")
        self._iterations = iterations
        self._distribution = self.Distribution.from_expr(distribution)
        self._minimum = self._freeze_recursive(minimum)
        self._maximum = self._freeze_recursive(maximum)

    def _iterate(self, state=None):
        def procedure(one, two):
            minimum, maximum = sorted([one, two])
            number = next(rng)
            return (number * (maximum - minimum)) + minimum

        rng = self._get_rng()
        for _ in self._loop(self._iterations):
            expr = self._apply_recursive(procedure, self._minimum, self._maximum)
            if (yield expr):
                return

    @property
    def distribution(self):
        return self._distribution

    @property
    def is_infinite(self):
        return self._iterations is None

    @property
    def iterations(self):
        return self._iterations

    @property
    def minimum(self):
        return self._minimum

    @property
    def maximum(self):
        return self._maximum


class ShufflePattern(SequencePattern):
    def __init__(self, sequence, iterations=1, forbid_repetitions=False):
        super().__init__(sequence, iterations=iterations)
        self._forbid_repetitions = bool(forbid_repetitions)

    def _iterate(self, state=None):
        should_stop = False
        rng = self._get_rng()
        previous_index = None
        for _ in self._loop(self._iterations):
            indices = self._shuffle(len(self._sequence), rng, previous_index)
            while (
                self.forbid_repetitions
                and len(indices) > 1
                and indices[0] == previous_index
            ):
                indices = self._shuffle(len(self._sequence), rng, previous_index)
            previous_index = indices[-1]
            for index in indices:
                choice = self._sequence[index]
                if isinstance(choice, Pattern):
                    should_stop = (yield from choice) or should_stop
                else:
                    should_stop = (yield choice) or should_stop
                if should_stop:
                    return

    def _shuffle(self, length, rng, previous_index=None):
        indices = list(range(length))
        shuffled_indices = []
        while len(indices) > 1:
            index = int(next(rng) * 0x7FFFFFFF) % len(indices)
            shuffled_indices.append(indices.pop(index))
        if indices:
            shuffled_indices.append(indices.pop())
        return shuffled_indices

    @property
    def forbid_repetitions(self):
        return self._forbid_repetitions



from typing import Optional, Union

from uqbar.objects import get_repr, new

from supriya.system import SupriyaValueObject


class Interval(SupriyaValueObject):
    """
    An interval (typically of time).

    ::

        >>> from supriya.intervals import Interval
        >>> Interval(0, 10)
        Interval(start_offset=0.0, stop_offset=10.0)

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_start_offset", "_stop_offset")

    ### INITIALIZER ###

    def __init__(
        self, start_offset: float = float("-inf"), stop_offset: float = float("inf")
    ):
        if start_offset is None:
            start_offset = float("-inf")
        if stop_offset is None:
            stop_offset = float("inf")
        start_offset, stop_offset = float(start_offset), float(stop_offset)
        assert start_offset <= stop_offset
        self._start_offset = start_offset
        self._stop_offset = stop_offset

    ### SPECIAL METHODS ###

    def __and__(self, interval: "Interval"):
        """
        Logical AND of two intervals:

        ::

            >>> from supriya.intervals import Interval
            >>> interval_1 = Interval(0, 10)
            >>> interval_2 = Interval(5, 12)
            >>> interval_3 = Interval(-2, 2)
            >>> interval_4 = Interval(10, 20)

        ::

            >>> interval_1 & interval_2
            IntervalTree(
                intervals=[
                    Interval(start_offset=5.0, stop_offset=10.0),
                ],
            )

        ::

            >>> interval_1 & interval_3
            IntervalTree(
                intervals=[
                    Interval(start_offset=0.0, stop_offset=2.0),
                ],
            )

        ::

            >>> interval_1 & interval_4
            IntervalTree(intervals=[])

        ::

            >>> interval_2 & interval_3
            IntervalTree(intervals=[])

        ::

            >>> interval_2 & interval_4
            IntervalTree(
                intervals=[
                    Interval(start_offset=10.0, stop_offset=12.0),
                ],
            )

        ::

            >>> interval_3 & interval_4
            IntervalTree(intervals=[])

        """
        from .IntervalTree import IntervalTree

        result = IntervalTree()
        if self.intersects(interval):
            new_interval = self.new(
                start_offset=max(self.start_offset, interval.start_offset),
                stop_offset=min(self.stop_offset, interval.stop_offset),
            )
            result.add(new_interval)
        return result

    def __contains__(self, interval: "Interval"):
        pass

    def __ge__(self, expr: Union["Interval", float]):
        if isinstance(expr, Interval):
            if self.start_offset > expr.start_offset:
                return True
            return (
                self.start_offset == expr.start_offset
                and self.stop_offset >= expr.stop_offset
            )
        return self.start_offset >= expr

    def __gt__(self, expr: Union["Interval", float]):
        if isinstance(expr, Interval):
            if self.start_offset > expr.start_offset:
                return True
            return (
                self.start_offset == expr.start_offset
                and self.stop_offset > expr.stop_offset
            )
        return self.start_offset > expr

    def __le__(self, expr: Union["Interval", float]):
        if isinstance(expr, Interval):
            if self.start_offset <= expr.start_offset:
                return True
            return (
                self.start_offset == expr.start_offset
                and self.stop_offset <= expr.stop_offset
            )
        return self.start_offset <= expr

    def __lt__(self, expr: Union["Interval", float]):
        if isinstance(expr, Interval):
            if self.start_offset < expr.start_offset:
                return True
            return (
                self.start_offset == expr.start_offset
                and self.stop_offset < expr.stop_offset
            )
        return self.start_offset < expr

    def __or__(self, interval: "Interval"):
        """
        Logical OR of two intervals:

        ::

            >>> from supriya.intervals import Interval
            >>> interval_1 = Interval(0, 10)
            >>> interval_2 = Interval(5, 12)
            >>> interval_3 = Interval(-2, 2)
            >>> interval_4 = Interval(10, 20)

        ::

            >>> interval_1 | interval_2
            IntervalTree(
                intervals=[
                    Interval(start_offset=0.0, stop_offset=12.0),
                ],
            )

        ::

            >>> interval_1 | interval_3
            IntervalTree(
                intervals=[
                    Interval(start_offset=-2.0, stop_offset=10.0),
                ],
            )

        ::

            >>> interval_1 | interval_4
            IntervalTree(
                intervals=[
                    Interval(start_offset=0.0, stop_offset=20.0),
                ],
            )

        ::

            >>> interval_2 | interval_3
            IntervalTree(
                intervals=[
                    Interval(start_offset=-2.0, stop_offset=2.0),
                    Interval(start_offset=5.0, stop_offset=12.0),
                ],
            )

        ::

            >>> interval_2 | interval_4
            IntervalTree(
                intervals=[
                    Interval(start_offset=5.0, stop_offset=20.0),
                ],
            )

        ::

            >>> interval_3 | interval_4
            IntervalTree(
                intervals=[
                    Interval(start_offset=-2.0, stop_offset=2.0),
                    Interval(start_offset=10.0, stop_offset=20.0),
                ],
            )

        """
        from .IntervalTree import IntervalTree

        result = IntervalTree()
        if (
            self.stop_offset < interval.start_offset
            or self.start_offset > interval.stop_offset
        ):
            result.update([self, interval])
            return result
        new_interval = self.new(
            start_offset=min(self.start_offset, interval.start_offset),
            stop_offset=max(self.stop_offset, interval.stop_offset),
        )
        result.add(new_interval)
        return result

    def __repr__(self):
        return get_repr(self, multiline=False)

    def __sub__(self, interval: "Interval"):
        """
        Subtract ``interval`` from interval:

        ::

            >>> from supriya.intervals import Interval
            >>> interval_1 = Interval(0, 10)
            >>> interval_2 = Interval(5, 12)
            >>> interval_3 = Interval(8, 11)
            >>> interval_4 = Interval(10, 20)

        ::

            >>> interval_1 - interval_1
            IntervalTree(intervals=[])

        ::

            >>> interval_1 - interval_2
            IntervalTree(
                intervals=[
                    Interval(start_offset=0.0, stop_offset=5.0),
                ],
            )

        ::

            >>> interval_1 - interval_3
            IntervalTree(
                intervals=[
                    Interval(start_offset=0.0, stop_offset=8.0),
                ],
            )

        ::

            >>> interval_1 - interval_4
            IntervalTree(
                intervals=[
                    Interval(start_offset=0.0, stop_offset=10.0),
                ],
            )

        ::

            >>> interval_2 - interval_1
            IntervalTree(
                intervals=[
                    Interval(start_offset=10.0, stop_offset=12.0),
                ],
            )

        ::

            >>> interval_2 - interval_2
            IntervalTree(intervals=[])

        ::

            >>> interval_2 - interval_3
            IntervalTree(
                intervals=[
                    Interval(start_offset=5.0, stop_offset=8.0),
                    Interval(start_offset=11.0, stop_offset=12.0),
                ],
            )

        ::

            >>> interval_2 - interval_4
            IntervalTree(
                intervals=[
                    Interval(start_offset=5.0, stop_offset=10.0),
                ],
            )

        ::

            >>> interval_3 - interval_1
            IntervalTree(
                intervals=[
                    Interval(start_offset=10.0, stop_offset=11.0),
                ],
            )

        ::

            >>> interval_3 - interval_2
            IntervalTree(intervals=[])

        ::

            >>> interval_3 - interval_3
            IntervalTree(intervals=[])

        ::

            >>> interval_3 - interval_4
            IntervalTree(
                intervals=[
                    Interval(start_offset=8.0, stop_offset=10.0),
                ],
            )

        ::

            >>> interval_4 - interval_1
            IntervalTree(
                intervals=[
                    Interval(start_offset=10.0, stop_offset=20.0),
                ],
            )

        ::

            >>> interval_4 - interval_2
            IntervalTree(
                intervals=[
                    Interval(start_offset=12.0, stop_offset=20.0),
                ],
            )

        ::

            >>> interval_4 - interval_3
            IntervalTree(
                intervals=[
                    Interval(start_offset=11.0, stop_offset=20.0),
                ],
            )

        ::

            >>> interval_4 - interval_4
            IntervalTree(intervals=[])

        """
        from .IntervalTree import IntervalTree

        result = IntervalTree()
        if self.start_offset < interval.start_offset:
            if self.stop_offset <= interval.start_offset:
                result.add(self)
            elif self.stop_offset <= interval.stop_offset:
                new_interval = self.new(stop_offset=interval.start_offset)
                result.add(new_interval)
            else:
                interval_one = self.new(stop_offset=interval.start_offset)
                interval_two = self.new(start_offset=interval.stop_offset)
                result.update([interval_one, interval_two])
        elif self.stop_offset > interval.stop_offset:
            new_interval = self.new(
                start_offset=max(self.start_offset, interval.stop_offset),
                stop_offset=self.stop_offset,
            )
            result.add(new_interval)
        return result

    def __xor__(self, interval: "Interval"):
        """
        Logical XOR of two intervals:

        ::

            >>> from supriya.intervals import Interval
            >>> interval_1 = Interval(0, 10)
            >>> interval_2 = Interval(5, 12)
            >>> interval_3 = Interval(-2, 2)
            >>> interval_4 = Interval(10, 20)

        ::

            >>> interval_1 ^ interval_2
            IntervalTree(
                intervals=[
                    Interval(start_offset=0.0, stop_offset=5.0),
                    Interval(start_offset=10.0, stop_offset=12.0),
                ],
            )

        ::

            >>> interval_1 ^ interval_3
            IntervalTree(
                intervals=[
                    Interval(start_offset=-2.0, stop_offset=0.0),
                    Interval(start_offset=2.0, stop_offset=10.0),
                ],
            )

        ::

            >>> interval_1 ^ interval_4
            IntervalTree(
                intervals=[
                    Interval(start_offset=0.0, stop_offset=10.0),
                    Interval(start_offset=10.0, stop_offset=20.0),
                ],
            )

        ::

            >>> interval_2 ^ interval_3
            IntervalTree(
                intervals=[
                    Interval(start_offset=-2.0, stop_offset=2.0),
                    Interval(start_offset=5.0, stop_offset=12.0),
                ],
            )

        ::

            >>> interval_2 ^ interval_4
            IntervalTree(
                intervals=[
                    Interval(start_offset=5.0, stop_offset=10.0),
                    Interval(start_offset=12.0, stop_offset=20.0),
                ],
            )

        ::

            >>> interval_3 ^ interval_4
            IntervalTree(
                intervals=[
                    Interval(start_offset=-2.0, stop_offset=2.0),
                    Interval(start_offset=10.0, stop_offset=20.0),
                ],
            )

        """
        from .IntervalTree import IntervalTree

        result = IntervalTree()
        if (
            self.stop_offset <= interval.start_offset
            or self.start_offset >= interval.stop_offset
        ):
            result.update([self, interval])
            return result
        starts = sorted([self.start_offset, interval.start_offset])
        stops = sorted([self.stop_offset, interval.stop_offset])
        if starts[0] < starts[1]:
            result.add(self.new(start_offset=starts[0], stop_offset=starts[1]))
        if stops[0] < stops[1]:
            result.add(self.new(start_offset=stops[0], stop_offset=stops[1]))
        return result

    ### PUBLIC METHODS ###

    def intersects(self, expr: Union["Interval", float]) -> bool:
        """
        True when interval intersects ``expr``, another interval or offset:

        ::

            >>> from supriya.intervals import Interval
            >>> interval_1 = Interval(0, 10)
            >>> interval_2 = Interval(5, 15)
            >>> interval_3 = Interval(10, 15)
            >>> interval_4 = Interval(15, 25)

        ::

            >>> interval_1.intersects(interval_1)
            True
            >>> interval_1.intersects(interval_2)
            True
            >>> interval_1.intersects(interval_3)
            False
            >>> interval_1.intersects(interval_4)
            False

        ::

            >>> for offset in [-5, 0, 5, 10, 15]:
            ...     print(offset, interval_1.intersects(offset))
            ...
            -5 False
            0 True
            5 True
            10 False
            15 False

        """
        if isinstance(expr, Interval):
            return (
                expr.start_offset <= self.start_offset
                and self.start_offset < expr.stop_offset
            ) or (
                self.start_offset <= expr.start_offset
                and expr.start_offset < self.stop_offset
            )
        return self.start_offset <= expr < self.stop_offset

    def is_tangent_to(self, expr: Union["Interval", float]) -> bool:
        """
        True when interval is tangent to ``expr``, another interval or offset:

        ::

            >>> from supriya.intervals import Interval
            >>> interval_1 = Interval(0, 10)
            >>> interval_2 = Interval(5, 15)
            >>> interval_3 = Interval(10, 15)
            >>> interval_4 = Interval(15, 25)

        ::

            >>> interval_1.is_tangent_to(interval_1)
            False
            >>> interval_1.is_tangent_to(interval_2)
            False
            >>> interval_1.is_tangent_to(interval_3)
            True
            >>> interval_1.is_tangent_to(interval_4)
            False

        ::

            >>> for offset in [-5, 0, 5, 10, 15]:
            ...     print(offset, interval_1.is_tangent_to(offset))
            ...
            -5 False
            0 True
            5 False
            10 True
            15 False

        """
        if isinstance(expr, Interval):
            return (
                self.start_offset == expr.stop_offset
                or self.stop_offset == expr.start_offset
            )
        return self.start_offset == expr or self.stop_offset == expr

    def new(self, start_offset=None, stop_offset=None, **kwargs):
        """
        Template a new interval:

        ::

            >>> from supriya.intervals import Interval
            >>> interval_one = Interval(0, 10)

            >>> interval_two = interval_one.new()
            >>> interval_two
            Interval(start_offset=0.0, stop_offset=10.0)
            >>> interval_two is not interval_one
            True

        ::

            >>> interval_two.new(start_offset=5.0)
            Interval(start_offset=5.0, stop_offset=10.0)

        """
        if start_offset is not None:
            kwargs["start_offset"] = start_offset
        if stop_offset is not None:
            kwargs["stop_offset"] = stop_offset
        return new(self, **kwargs)

    def split(self, *offsets):
        """
        Split at ``offsets``:

        ::

            >>> from supriya.intervals import Interval
            >>> interval = Interval(0, 10)
            >>> for split_interval in interval.split(1, 3, 7):
            ...     split_interval
            ...
            Interval(start_offset=0.0, stop_offset=1.0)
            Interval(start_offset=1.0, stop_offset=3.0)
            Interval(start_offset=3.0, stop_offset=7.0)
            Interval(start_offset=7.0, stop_offset=10.0)

        """
        from .IntervalTree import IntervalTree

        split_offsets = sorted(
            float(offset)
            for offset in offsets
            if self._start_offset < offset < self._stop_offset
        )
        result = IntervalTree()
        right_interval = self
        for offset in split_offsets:
            left_interval = new(right_interval, stop_offset=offset)
            right_interval = new(right_interval, start_offset=offset)
            result.add(left_interval)
        result.add(right_interval)
        return result

    def translate(self, translation: float, stop_translation: Optional[float] = None):
        """
        Translate offsets by ``translation``, and optionally translate independently when
        specifying ``stop_translation``:

        ::

            >>> from supriya.intervals import Interval
            >>> interval = Interval(0, 10)

        ::

            >>> interval.translate(5)
            Interval(start_offset=5.0, stop_offset=15.0)

        ::

            >>> interval.translate(5, 7.5)
            Interval(start_offset=5.0, stop_offset=17.5)

        """

        start_translation = float(translation)
        stop_translation = (
            float(stop_translation)
            if stop_translation is not None
            else start_translation
        )
        return self.new(
            start_offset=self._start_offset + start_translation,
            stop_offset=self._stop_offset + stop_translation,
        )

    ### PUBLIC PROPERTIES ###

    @property
    def duration(self):
        return self.size

    @property
    def size(self) -> float:
        return self._stop_offset - self._start_offset

    @property
    def start_offset(self) -> float:
        return self._start_offset

    @property
    def stop_offset(self) -> float:
        return self._stop_offset

    @property
    def wellformed(self) -> bool:
        return self._start_offset < self._stop_offset



import threading
from typing import Set

from uqbar.objects import new

from supriya.intervals import IntervalTree
from supriya.intervals.Interval import Interval
from supriya.system import SupriyaObject


class Block(Interval):

    ### CLASS VARIABLES ###

    __slots__ = ("_used",)

    ### INITIALIZER ###

    def __init__(
        self,
        start_offset: float = float("-inf"),
        stop_offset: float = float("inf"),
        used: bool = False,
    ):
        Interval.__init__(self, start_offset=start_offset, stop_offset=stop_offset)
        self._used = bool(used)

    ### PUBLIC PROPERTIES ###

    @property
    def used(self) -> bool:
        return self._used


class BlockAllocator(SupriyaObject):
    """
    A block allocator.

    ::

        >>> from supriya.realtime import BlockAllocator
        >>> allocator = BlockAllocator(
        ...     heap_maximum=16,
        ... )

    ::

        >>> allocator.allocate(4)
        0

    ::

        >>> allocator.allocate(4)
        4

    ::

        >>> allocator.allocate(4)
        8

    ::

        >>> allocator.allocate(8) is None
        True

    ::

        >>> allocator.free(8)
        >>> allocator.allocate(8)
        8

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_free_heap", "_heap_maximum", "_heap_minimum", "_lock", "_used_heap")

    ### INITIALIZER ###

    def __init__(self, heap_maximum=None, heap_minimum=0):
        import supriya.realtime

        self._free_heap = IntervalTree(accelerated=True)
        self._heap_maximum = heap_maximum
        self._heap_minimum = heap_minimum
        self._lock = threading.Lock()
        self._used_heap = IntervalTree(accelerated=True)
        free_block = supriya.realtime.Block(
            start_offset=heap_minimum, stop_offset=heap_maximum, used=False
        )
        self._free_heap.add(free_block)

    ### PUBLIC METHODS ###

    def allocate(self, desired_block_size=1):
        desired_block_size = int(desired_block_size)
        assert 0 < desired_block_size
        block_id = None
        with self._lock:
            free_block = None
            for block in self._free_heap:
                if desired_block_size <= block.duration:
                    free_block = block
                    break
            if free_block is not None:
                split_offset = free_block.start_offset + desired_block_size
                self._free_heap.remove(free_block)
                if desired_block_size < free_block.duration:
                    new_free_block = new(
                        free_block, start_offset=split_offset, used=False
                    )
                    self._free_heap.add(new_free_block)
                    used_block = new(free_block, stop_offset=split_offset, used=True)
                else:
                    used_block = new(free_block, used=True)
                self._used_heap.add(used_block)
                block_id = used_block.start_offset
        if block_id is None:
            return block_id
        return int(block_id)

    def allocate_at(self, index=None, desired_block_size=1):
        import supriya.realtime

        index = int(index)
        desired_block_size = int(desired_block_size)
        block_id = None
        with self._lock:
            start_offset = index
            stop_offset = index + desired_block_size
            start_cursor = self._free_heap.get_moment_at(start_offset)
            starting_blocks = sorted(
                start_cursor.start_intervals + start_cursor.overlap_intervals
            )
            stop_cursor = self._free_heap.get_moment_at(stop_offset)
            stop_blocks = sorted(
                stop_cursor.overlap_intervals + stop_cursor.stop_intervals
            )
            if starting_blocks == stop_blocks:
                assert len(starting_blocks) == 1
                free_block = starting_blocks[0]
                used_block = supriya.realtime.Block(
                    start_offset=start_offset, stop_offset=stop_offset, used=True
                )
                self._used_heap.add(used_block)
                self._free_heap.remove(free_block)
                free_blocks = free_block - used_block
                for free_block in free_blocks:
                    self._free_heap.add(free_block)
                block_id = index
        if block_id is None:
            return block_id
        return int(block_id)

    def free(self, block_id):
        import supriya.realtime

        block_id = int(block_id)
        with self._lock:
            cursor = self._used_heap.get_moment_at(block_id)
            blocks = sorted(
                set(cursor.start_intervals) or set(cursor.overlap_intervals)
            )
            assert len(blocks) == 1
            used_block = blocks[0]
            self._used_heap.remove(used_block)
            start_offset = used_block.start_offset
            stopping_blocks = self._free_heap.find_intervals_stopping_at(start_offset)
            if stopping_blocks:
                assert len(stopping_blocks) == 1
                stopping_block = stopping_blocks[0]
                self._free_heap.remove(stopping_block)
                start_offset = stopping_block.start_offset
            stop_offset = used_block.stop_offset
            starting_blocks = self._free_heap.find_intervals_starting_at(stop_offset)
            if starting_blocks:
                assert len(starting_blocks) == 1
                starting_block = starting_blocks[0]
                self._free_heap.remove(starting_block)
                stop_offset = starting_block.stop_offset
            free_block = supriya.realtime.Block(
                start_offset=start_offset, stop_offset=stop_offset, used=False
            )
            self._free_heap.add(free_block)

    ### PUBLIC PROPERTIES ###

    @property
    def heap_maximum(self):
        """
        Maximum allocatable index.
        """
        return self._heap_maximum

    @property
    def heap_minimum(self):
        """
        Minimum allocatable index.
        """
        return self._heap_minimum


class NodeIdAllocator(SupriyaObject):
    """
    A node ID allocator.

    ::

        >>> from supriya.realtime import NodeIdAllocator
        >>> allocator = NodeIdAllocator()
        >>> for _ in range(3):
        ...     allocator.allocate_node_id()
        ...
        1000
        1001
        1002

    ::

        >>> for _ in range(3):
        ...     allocator.allocate_permanent_node_id()
        ...
        1
        2
        3

    ::

        >>> allocator.free_permanent_node_id(2)

    ::

        >>> allocator.allocate_permanent_node_id()
        2

    """

    ### INITIALIZER ###

    def __init__(self, client_id: int = 0, initial_node_id: int = 1000):
        if client_id > 31:
            raise ValueError
        self._initial_node_id = initial_node_id
        self._client_id = client_id
        self._mask = self._client_id << 26
        self._temp = self._initial_node_id
        self._next_permanent_id = 1
        self._freed_permanent_ids: Set[int] = set()
        self._lock = threading.Lock()

    ### PUBLIC METHODS ###

    def allocate_node_id(self, count: int = 1) -> int:
        with self._lock:
            x = self._temp
            temp = x + count
            if 0x03FFFFFF < temp:
                temp = (temp % 0x03FFFFFF) + self._initial_node_id
            self._temp = temp
            x = x | self._mask
        return x

    def allocate_permanent_node_id(self) -> int:
        with self._lock:
            if self._freed_permanent_ids:
                x = min(self._freed_permanent_ids)
                self._freed_permanent_ids.remove(x)
            else:
                x = self._next_permanent_id
                self._next_permanent_id = min(x + 1, self._initial_node_id - 1)
            x = x | self._mask
        return x

    def free_permanent_node_id(self, node_id: int) -> None:
        with self._lock:
            node_id = node_id & 0x03FFFFFF
            if node_id < self._initial_node_id:
                self._freed_permanent_ids.add(node_id)



import collections
from collections.abc import Sequence

from supriya import CalculationRate
from supriya.synthdefs import MultiOutUGen, UGen


class DUGen(UGen):
    """
    Abstract base class of demand-rate UGens.
    """

    def __init__(self, **kwargs):
        kwargs["calculation_rate"] = CalculationRate.DEMAND
        UGen.__init__(self, **kwargs)


class Dbrown(DUGen):
    """
    A demand-rate brownian movement generator.

    ::

        >>> dbrown = supriya.ugens.Dbrown.new(
        ...     length=float("inf"),
        ...     maximum=1,
        ...     minimum=0,
        ...     step=0.01,
        ... )
        >>> dbrown
        Dbrown()

    """

    _ordered_input_names = collections.OrderedDict(
        [("minimum", 0.0), ("maximum", 1.0), ("step", 0.01), ("length", float("inf"))]
    )
    _valid_calculation_rates = (CalculationRate.DEMAND,)


class Dbufrd(DUGen):
    """
    A buffer-reading demand-rate UGen.

    ::

        >>> dbufrd = supriya.ugens.Dbufrd(
        ...     buffer_id=0,
        ...     loop=1,
        ...     phase=0,
        ... )
        >>> dbufrd
        Dbufrd()

    """

    _ordered_input_names = collections.OrderedDict(
        [("buffer_id", 0), ("phase", 0), ("loop", 1)]
    )
    _valid_calculation_rates = (CalculationRate.DEMAND,)


class Dbufwr(DUGen):
    """
    A buffer-writing demand-rate UGen.

    ::

        >>> dbufwr = supriya.ugens.Dbufwr(
        ...     buffer_id=0,
        ...     source=0,
        ...     loop=1,
        ...     phase=0,
        ... )
        >>> dbufwr
        Dbufwr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", 0.0), ("buffer_id", 0.0), ("phase", 0.0), ("loop", 1.0)]
    )
    _valid_calculation_rates = (CalculationRate.DEMAND,)


class Demand(MultiOutUGen):
    """
    Demands results from demand-rate UGens.

    ::

        >>> source = [
        ...     supriya.ugens.Dseries(),
        ...     supriya.ugens.Dwhite(),
        ... ]
        >>> trigger = supriya.ugens.Impulse.kr(1)
        >>> demand = supriya.ugens.Demand.ar(
        ...     reset=0,
        ...     source=source,
        ...     trigger=trigger,
        ... )
        >>> demand
        UGenArray({2})

    """

    ### CLASS VARIABLES ###

    _default_channel_count = 1
    _has_settable_channel_count = False
    _ordered_input_names = collections.OrderedDict(
        [("trigger", 0), ("reset", 0), ("source", None)]
    )
    _unexpanded_input_names = ("source",)
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)

    ### INITIALIZER ###

    def __init__(self, calculation_rate=None, trigger=None, reset=None, source=None):
        if not isinstance(source, Sequence):
            source = [source]
        MultiOutUGen.__init__(
            self,
            calculation_rate=calculation_rate,
            trigger=trigger,
            reset=reset,
            source=source,
            channel_count=len(source),
        )


class DemandEnvGen(UGen):
    """
    A demand rate envelope generator.

    ::

        >>> demand_env_gen = supriya.ugens.DemandEnvGen.ar(
        ...     curve=0,
        ...     done_action=0,
        ...     duration=1,
        ...     gate=1,
        ...     level=1,
        ...     level_bias=0,
        ...     level_scale=1,
        ...     reset=1,
        ...     shape=1,
        ...     time_scale=1,
        ... )
        >>> demand_env_gen
        DemandEnvGen.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("level", None),
            ("duration", None),
            ("shape", 1),
            ("curve", 0),
            ("gate", 1),
            ("reset", 1),
            ("level_scale", 1),
            ("level_bias", 0),
            ("time_scale", 1),
            ("done_action", 0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Dgeom(DUGen):
    """
    A demand-rate geometric series generator.

    ::

        >>> dgeom = supriya.ugens.Dgeom.new(
        ...     grow=2,
        ...     length=float("inf"),
        ...     start=1,
        ... )
        >>> dgeom
        Dgeom()

    """

    _ordered_input_names = collections.OrderedDict(
        [("start", 1), ("grow", 2), ("length", float("inf"))]
    )
    _valid_calculation_rates = (CalculationRate.DEMAND,)


class Dibrown(DUGen):
    """
    An integer demand-rate brownian movement generator.

    ::

        >>> dibrown = supriya.ugens.Dibrown.new(
        ...     length=float("inf"),
        ...     maximum=1,
        ...     minimum=0,
        ...     step=0.01,
        ... )
        >>> dibrown
        Dibrown()

    """

    _ordered_input_names = collections.OrderedDict(
        [("minimum", 0), ("maximum", 12), ("step", 1), ("length", float("inf"))]
    )
    _valid_calculation_rates = (CalculationRate.DEMAND,)


class Diwhite(DUGen):
    """
    An integer demand-rate white noise random generator.

    ::

        >>> diwhite = supriya.ugens.Diwhite.new(
        ...     length=float("inf"),
        ...     maximum=1,
        ...     minimum=0,
        ... )
        >>> diwhite
        Diwhite()

    """

    _ordered_input_names = collections.OrderedDict(
        [("minimum", 0), ("maximum", 1), ("length", float("inf"))]
    )
    _valid_calculation_rates = (CalculationRate.DEMAND,)


class Drand(DUGen):
    """
    A demand-rate random sequence generator.

    ::

        >>> sequence = (1, 2, 3)
        >>> drand = supriya.ugens.Drand.new(
        ...     repeats=1,
        ...     sequence=sequence,
        ... )
        >>> drand
        Drand()

    """

    _ordered_input_names = collections.OrderedDict([("repeats", 1), ("sequence", None)])
    _unexpanded_input_names = ("sequence",)
    _valid_calculation_rates = (CalculationRate.DEMAND,)


class Dreset(DUGen):
    """
    Resets demand-rate UGens.

    ::

        >>> source = supriya.ugens.Dseries(start=0, step=2)
        >>> dreset = supriya.ugens.Dreset(
        ...     reset=0,
        ...     source=source,
        ... )
        >>> dreset
        Dreset()

    """

    _ordered_input_names = collections.OrderedDict([("source", None), ("reset", 0)])
    _valid_calculation_rates = (CalculationRate.DEMAND,)


class Dseq(DUGen):
    """
    A demand-rate sequence generator.

    ::

        >>> sequence = (1, 2, 3)
        >>> dseq = supriya.ugens.Dseq.new(
        ...     repeats=1,
        ...     sequence=sequence,
        ... )
        >>> dseq
        Dseq()

    """

    _ordered_input_names = collections.OrderedDict([("repeats", 1), ("sequence", None)])
    _unexpanded_input_names = ("sequence",)
    _valid_calculation_rates = (CalculationRate.DEMAND,)


class Dser(DUGen):
    """
    A demand-rate sequence generator.

    ::

        >>> sequence = (1, 2, 3)
        >>> dser = supriya.ugens.Dser.new(
        ...     repeats=1,
        ...     sequence=sequence,
        ... )
        >>> dser
        Dser()

    """

    _ordered_input_names = collections.OrderedDict([("repeats", 1), ("sequence", None)])
    _unexpanded_input_names = ("sequence",)
    _valid_calculation_rates = (CalculationRate.DEMAND,)


class Dseries(DUGen):
    """
    A demand-rate arithmetic series.

    ::

        >>> dseries = supriya.ugens.Dseries.new(
        ...     length=float("inf"),
        ...     start=1,
        ...     step=1,
        ... )
        >>> dseries
        Dseries()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict(
        [("start", 1), ("step", 1), ("length", float("inf"))]
    )
    _valid_calculation_rates = (CalculationRate.DEMAND,)

    ### INITIALIZER ###

    def __init__(self, calculation_rate=None, length=float("inf"), start=1, step=1):
        if length is None:
            length = float("inf")
        DUGen.__init__(self, length=length, start=start, step=step)


class Dshuf(DUGen):
    """
    A demand-rate random sequence generator.

    ::

        >>> sequence = (1, 2, 3)
        >>> dshuf = supriya.ugens.Dshuf.new(
        ...     repeats=1,
        ...     sequence=sequence,
        ... )
        >>> dshuf
        Dshuf()

    """

    _ordered_input_names = collections.OrderedDict([("repeats", 1), ("sequence", None)])
    _unexpanded_input_names = ("sequence",)
    _valid_calculation_rates = (CalculationRate.DEMAND,)


class Dstutter(DUGen):
    """
    A demand-rate input replicator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> dstutter = supriya.ugens.Dstutter.new(
        ...     n=2,
        ...     source=source,
        ... )
        >>> dstutter
        Dstutter()

    """

    _ordered_input_names = collections.OrderedDict([("n", 2.0), ("source", None)])
    _valid_calculation_rates = (CalculationRate.DEMAND,)


class Dswitch(DUGen):
    """
    A demand-rate generator for embedding different inputs.

    ::

        >>> index = supriya.ugens.Dseq(sequence=[0, 1, 2, 1, 0])
        >>> sequence = (1.0, 2.0, 3.0)
        >>> dswitch = supriya.ugens.Dswitch.new(
        ...     index=index,
        ...     sequence=sequence,
        ... )
        >>> dswitch
        Dswitch()

    """

    _ordered_input_names = collections.OrderedDict(
        [("index", None), ("sequence", None)]
    )
    _unexpanded_input_names = ("sequence",)
    _valid_calculation_rates = (CalculationRate.DEMAND,)


class Dswitch1(DUGen):
    """
    A demand-rate generator for switching between inputs.

    ::

        >>> index = supriya.ugens.Dseq(sequence=[0, 1, 2, 1, 0])
        >>> sequence = (1.0, 2.0, 3.0)
        >>> dswitch_1 = supriya.ugens.Dswitch1.new(
        ...     index=index,
        ...     sequence=sequence,
        ... )
        >>> dswitch_1
        Dswitch1()

    """

    _ordered_input_names = collections.OrderedDict(
        [("index", None), ("sequence", None)]
    )
    _unexpanded_input_names = ("sequence",)
    _valid_calculation_rates = (CalculationRate.DEMAND,)


class Dunique(DUGen):
    """
    Returns the same unique series of values for several demand streams.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> dunique = supriya.ugens.Dunique.new(
        ...     max_buffer_size=1024,
        ...     protected=True,
        ...     source=source,
        ... )
        >>> dunique
        Dunique()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("max_buffer_size", 1024), ("protected", True)]
    )
    _valid_calculation_rates = (CalculationRate.DEMAND,)


class Duty(UGen):
    """
    A value is demanded of each UGen in the list and output according to a stream of duration values.

    ::

        >>> duty = supriya.ugens.Duty.kr(
        ...     done_action=0,
        ...     duration=supriya.ugens.Drand(
        ...         sequence=[0.01, 0.2, 0.4],
        ...         repeats=2,
        ...     ),
        ...     reset=0,
        ...     level=supriya.ugens.Dseq(
        ...         sequence=[204, 400, 201, 502, 300, 200],
        ...         repeats=2,
        ...     ),
        ... )
        >>> duty
        Duty.kr()

    """

    _ordered_input_names = collections.OrderedDict(
        [("duration", 1.0), ("reset", 0.0), ("level", 1.0), ("done_action", 0.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Dwhite(DUGen):
    """
    A demand-rate white noise random generator.

    ::

        >>> dwhite = supriya.ugens.Dwhite.new(
        ...     length=float("inf"),
        ...     maximum=1,
        ...     minimum=0,
        ... )
        >>> dwhite
        Dwhite()

    """

    _ordered_input_names = collections.OrderedDict(
        [("minimum", 0.0), ("maximum", 1.0), ("length", float("inf"))]
    )
    _valid_calculation_rates = (CalculationRate.DEMAND,)


class Dwrand(DUGen):
    """
    A demand-rate weighted random sequence generator.

    ::

        >>> sequence = [0, 1, 2, 7]
        >>> weights = [0.4, 0.4, 0.1, 0.1]
        >>> dwrand = supriya.ugens.Dwrand.new(
        ...     repeats=1,
        ...     sequence=sequence,
        ...     weights=weights,
        ... )
        >>> dwrand
        Dwrand()

    """

    ### CLASS VARIABLES ###

    # TODO: We should not include length in the generated methods

    _ordered_input_names = collections.OrderedDict(
        [("repeats", 1), ("length", None), ("weights", None), ("sequence", None)]
    )
    _unexpanded_input_names = ("weights", "sequence")
    _valid_calculation_rates = (CalculationRate.DEMAND,)

    ### INITIALIZER ###

    def __init__(self, repeats=1, sequence=None, weights=None, **kwargs):
        if not isinstance(sequence, Sequence):
            sequence = [sequence]
        sequence = tuple(float(_) for _ in sequence)
        if not isinstance(weights, Sequence):
            weights = [weights]
        weights = tuple(float(_) for _ in weights)
        weights = weights[: len(sequence)]
        weights += (0.0,) * (len(sequence) - len(weights))
        DUGen.__init__(
            self,
            repeats=repeats,
            length=len(sequence),
            sequence=sequence,
            weights=weights,
        )


class Dxrand(DUGen):
    """
    A demand-rate random sequence generator.

    ::

        >>> sequence = (1, 2, 3)
        >>> dxrand = supriya.ugens.Dxrand.new(
        ...     repeats=1,
        ...     sequence=sequence,
        ... )
        >>> dxrand
        Dxrand()

    """

    _ordered_input_names = collections.OrderedDict([("repeats", 1), ("sequence", None)])
    _unexpanded_input_names = ("sequence",)
    _valid_calculation_rates = (CalculationRate.DEMAND,)



"""
Tools for interacting with and modeling objects on the SuperCollider
``scsynth`` synthesis server.
"""
from .allocators import Block, BlockAllocator, NodeIdAllocator
from .bases import ServerObject
from .buffers import Buffer, BufferGroup, BufferProxy
from .buses import AudioInputBusGroup, AudioOutputBusGroup, Bus, BusGroup, BusProxy
from .interfaces import (
    ControlInterface,
    GroupControl,
    GroupInterface,
    SynthControl,
    SynthInterface,
)
from .meters import Meters
from .nodes import Group, Node, RootNode, Synth
from .recorder import Recorder
from .servers import AsyncServer, BaseServer, Server

__all__ = [
    "AudioInputBusGroup",
    "AudioOutputBusGroup",
    "AsyncServer",
    "BaseServer",
    "Block",
    "BlockAllocator",
    "Buffer",
    "BufferGroup",
    "BufferProxy",
    "Bus",
    "BusGroup",
    "BusProxy",
    "ControlInterface",
    "Group",
    "GroupControl",
    "GroupInterface",
    "Meters",
    "Node",
    "NodeIdAllocator",
    "Recorder",
    "RootNode",
    "Server",
    "ServerObject",
    "Synth",
    "SynthControl",
    "SynthInterface",
    "boot",
]



import abc
import pathlib
import tempfile
from collections.abc import Sequence
from typing import Optional, Tuple, cast

import uqbar.graphs
import uqbar.strings
from uqbar.containers import UniqueTreeList, UniqueTreeNode
from uqbar.objects import new

import supriya
from supriya.enums import AddAction, NodeAction
from supriya.exceptions import ServerOffline

from ..synthdefs.synthdefs import SynthDef
from ..typing import AddActionLike
from .interfaces import GroupInterface, SynthInterface  # noqa


class Node(UniqueTreeNode):

    ### CLASS VARIABLES ###

    _valid_add_actions: Tuple[int, ...] = ()

    ### INITIALIZER ###

    @abc.abstractmethod
    def __init__(self, name=None, node_id_is_permanent=False):
        UniqueTreeNode.__init__(self, name=name)
        self._is_paused = False
        self._node_id = None
        self._node_id_is_permanent = bool(node_id_is_permanent)
        self._server = None

    ### SPECIAL METHODS ###

    def __float__(self):
        return float(self.node_id)

    def __graph__(self):
        graph = uqbar.graphs.Graph(
            attributes={
                "bgcolor": "transparent",
                "color": "lightslategrey",
                "dpi": 72,
                "fontname": "Arial",
                "outputorder": "edgesfirst",
                "overlap": "prism",
                "penwidth": 2,
                "rankdir": "TB",
                "ranksep": 0.5,
                "splines": "spline",
                "style": ("dotted", "rounded"),
            },
            edge_attributes={"penwidth": 2},
            node_attributes={
                "fontname": "Arial",
                "fontsize": 12,
                "penwidth": 2,
                "shape": "Mrecord",
                "style": ("filled", "rounded"),
            },
            children=[self._as_graphviz_node()],
        )
        return graph

    def __int__(self):
        return int(self.node_id)

    def __repr__(self):
        class_name = type(self).__name__
        node_id = "???"
        if self.node_id is not None:
            node_id = self.node_id
        allocated = "+" if self.is_allocated else "-"
        if self.name is None:
            return f"<{allocated} {class_name}: {node_id}>"
        return "<{allocated} {class_name}: {node_id} ({name})>"

    ### PRIVATE METHODS ###

    def _allocate(self, paused_nodes, requests, server, synthdefs):
        import supriya.commands

        if paused_nodes:
            requests.append(
                supriya.commands.NodeRunRequest(
                    node_id_run_flag_pairs=[(node, False) for node in paused_nodes]
                )
            )
        if not requests:
            return self
        elif 1 < len(requests):
            request = supriya.commands.RequestBundle(contents=requests)
        else:
            request = requests[0]
        requests[:] = [request]
        if synthdefs:
            synthdef_request = supriya.commands.SynthDefReceiveRequest(
                synthdefs=synthdefs, callback=requests[0]
            )
            if len(synthdef_request.to_datagram(with_placeholders=True)) > 8192:
                directory_path = pathlib.Path(tempfile.mkdtemp())
                synthdef_request = supriya.commands.SynthDefLoadDirectoryRequest(
                    directory_path=directory_path, callback=requests[0]
                )
                for synthdef in synthdefs:
                    file_name = "{}.scsyndef".format(synthdef.anonymous_name)
                    synthdef_path = directory_path / file_name
                    synthdef_path.write_bytes(synthdef.compile())
            requests[:] = [synthdef_request]
            if len(requests[0].to_datagram(with_placeholders=True)) > 8192:
                node_allocate_request = requests[0].callback
                synthdef_request = new(requests[0], callback=None)
                requests[:] = [node_allocate_request, synthdef_request]
        for request in requests:
            request.communicate(server=server, sync=True)
        return self

    def _as_graphviz_node(self):
        node = uqbar.graphs.Node(name=self._get_graphviz_name())
        group = uqbar.graphs.RecordGroup()
        group.append(uqbar.graphs.RecordField(label=type(self).__name__))
        if self.name:
            group.append(uqbar.graphs.RecordField(label="name: " + self.name))
        if self.node_id is not None:
            label = "id: " + str(self.node_id)
        else:
            label = "id: " + "-".join(str(_) for _ in (0,) + self.graph_order)
        group.append(uqbar.graphs.RecordField(label=label))
        node.append(group)
        return node

    def _cache_control_interface(self):
        return self._control_interface.as_dict()

    @staticmethod
    def _expr_as_target(expr):
        from supriya import Server

        if isinstance(expr, Server):
            expr = expr.default_group
        if not isinstance(expr, Node):
            raise ValueError
        if not expr.is_allocated:
            raise ServerOffline
        return expr

    def _get_graphviz_name(self):
        parts = [uqbar.strings.to_dash_case(type(self).__name__)]
        if self.is_allocated:
            parts.append(str(self.node_id))
        else:
            parts.extend(str(_) for _ in self.graph_order)
        return "-".join(parts)

    def _handle_response(self, response):
        import supriya.commands

        if not isinstance(response, supriya.commands.NodeInfoResponse):
            return
        if response.action == NodeAction.NODE_REMOVED:
            self._set_parent(None)
            self._unregister_with_local_server()
        elif response.action == NodeAction.NODE_ACTIVATED:
            self._is_paused = False
        elif response.action == NodeAction.NODE_DEACTIVATED:
            self._is_paused = True
        elif response.action == NodeAction.NODE_MOVED:
            old_parent = self.parent
            new_parent = self.server._nodes[response.parent_id]
            old_index = self.parent.index(self)
            new_index = None
            previous_node = self.server._nodes.get(response.previous_node_id)
            if previous_node is not None:
                try:
                    new_index = new_parent.index(previous_node) + 1
                except ValueError:
                    pass
            next_node = self.server._nodes.get(response.next_node_id)
            if next_node is not None:
                try:
                    new_index = new_parent.index(next_node)
                except ValueError:
                    pass
            if new_parent is old_parent:
                if new_index is not None and new_index != old_index:
                    if new_index < old_index:
                        self.parent._children.remove(self)
                        self.parent._children.insert(new_index, self)
                    else:
                        self.parent._children.insert(new_index, self)
                        self.parent._children.pop(old_index)
            else:
                self._set_parent(new_parent)
                new_index = new_index or 0
                new_parent._children.insert(new_index, self)

    def _move_node(self, *, add_action, node):
        target_node = self
        if add_action in (AddAction.ADD_TO_HEAD, AddAction.ADD_TO_TAIL):
            parent_node = target_node
        else:
            parent_node = target_node._parent
        node._set_parent(parent_node)
        if add_action == AddAction.ADD_TO_HEAD:
            parent_node._children.insert(0, node)
        elif add_action == AddAction.ADD_TO_TAIL:
            parent_node._children.append(node)
        elif add_action == AddAction.ADD_BEFORE:
            index = parent_node.index(target_node)
            parent_node._children.insert(index, node)
        elif add_action == AddAction.ADD_AFTER:
            index = parent_node.index(target_node)
            parent_node._children.insert(index + 1, node)
        elif add_action == AddAction.REPLACE:
            index = parent_node.index(target_node)
            parent_node._children[index] = node
            target_node._set_parent(None)
            target_node._unregister_with_local_server()

    def _register_with_local_server(
        self,
        server: "supriya.realtime.servers.Server",
        node_id: Optional[int] = None,
        node_id_is_permanent: bool = False,
    ):
        id_allocator = server.node_id_allocator
        if node_id is None:
            if node_id_is_permanent:
                node_id = id_allocator.allocate_permanent_node_id()
            else:
                node_id = server.node_id_allocator.allocate_node_id()
        else:
            node_id = int(node_id)
        if not self.is_allocated or self.server is not server:
            self._server = server
        self._node_id = node_id
        self._node_id_is_permanent = bool(node_id_is_permanent)
        self._server._nodes[self._node_id] = self
        return node_id

    def _remove_control_interface_from_parentage(self, old_parent, name_dictionary):
        if old_parent is None or not name_dictionary:
            return
        for parent in old_parent.parentage:
            parent._control_interface.remove_controls(name_dictionary)

    def _restore_control_interface_to_parentage(self, new_parent, name_dictionary):
        if new_parent is None or not name_dictionary:
            return
        for parent in new_parent.parentage:
            parent._control_interface.add_controls(name_dictionary)

    def _run(self, run_flag):
        self._is_paused = not bool(run_flag)

    def _set_parent(self, new_parent):
        old_parent = self._parent
        named_children = self._cache_named_children()
        control_interface = self._cache_control_interface()
        self._remove_from_parent()
        self._remove_named_children_from_parentage(old_parent, named_children)
        self._remove_control_interface_from_parentage(old_parent, control_interface)
        self._parent = new_parent
        self._restore_named_children_to_parentage(new_parent, named_children)
        self._restore_control_interface_to_parentage(new_parent, control_interface)

    def _unregister_with_local_server(self):
        node_id = self.node_id
        if self.server is not None:
            if self._node_id in self._server._nodes:
                del self._server._nodes[self._node_id]
            if self.node_id_is_permanent:
                self.server.node_id_allocator.free_permanent_node_id(self.node_id)
        self._node_id = None
        self._node_id_is_permanent = None
        self._server = None
        return node_id

    ### PUBLIC METHODS ###

    def add_group(self, add_action: AddActionLike = None) -> "Group":
        """
        Add a group relative to this node via ``add_action``.

        ::

            >>> server = supriya.Server().boot()
            >>> print(server.query())
            NODE TREE 0 group
                1 group

        ::

            >>> node = server.add_group()
            >>> group = node.add_group()
            >>> print(server.query())
            NODE TREE 0 group
                1 group
                    1000 group
                        1001 group

        """
        if add_action is None:
            add_action = self._valid_add_actions[0]
        add_action = AddAction.from_expr(add_action)
        if add_action not in self._valid_add_actions:
            raise ValueError("Invalid add action: {add_action}")
        group = Group()
        group.allocate(add_action=add_action, target_node=self)
        return group

    def add_synth(
        self,
        synthdef: Optional[SynthDef] = None,
        add_action: AddActionLike = None,
        **kwargs,
    ) -> "Synth":
        """
        Add a synth relative to this node via ``add_action``.

        ::

            >>> server = supriya.Server().boot()
            >>> print(server.query())
            NODE TREE 0 group
                1 group

        ::

            >>> node = server.add_group()
            >>> synth = node.add_synth()
            >>> print(server.query())
            NODE TREE 0 group
                1 group
                    1000 group
                        1001 default
                            out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5

        """
        if add_action is None:
            add_action = self._valid_add_actions[0]
        add_action = AddAction.from_expr(add_action)
        if add_action not in self._valid_add_actions:
            raise ValueError("Invalid add action: {add_action}")
        synth = Synth(synthdef=synthdef, **kwargs)
        synth.allocate(add_action=add_action, target_node=self)
        return synth

    def free(self) -> "Node":
        import supriya.commands

        self._set_parent(None)
        server = self.server
        if self.node_id is not None and server is not None and server.is_running:
            node_id = self._unregister_with_local_server()
            node_free_request = supriya.commands.NodeFreeRequest(node_ids=(node_id,))
            node_free_request.communicate(server=server, sync=False)
        return self

    def move_node(self, node: "Node", add_action: AddActionLike = None) -> "Node":
        """
        Move ``node`` relative to this node via ``add_action``.

        ::

            >>> server = supriya.Server().boot()
            >>> group_one = server.add_group()
            >>> group_two = server.add_group()
            >>> print(server.query())
            NODE TREE 0 group
                1 group
                    1001 group
                    1000 group

        ::

            >>> _ = group_two.move_node(group_one, "add_to_head")
            >>> print(server.query())
            NODE TREE 0 group
                1 group
                    1001 group
                        1000 group

        """
        if add_action is None:
            add_action = self._valid_add_actions[0]
        add_action = AddAction.from_expr(add_action)
        if add_action not in self._valid_add_actions:
            raise ValueError("Invalid add action: {add_action}")
        elif node in self.parentage:
            raise ValueError("Node in parentage")
        if add_action == AddAction.ADD_BEFORE:
            if self.parent is None:
                raise ValueError("Cannot move before without parent")
            index = self.parent.index(self)
            self.parent.insert(index, node)
        elif add_action == AddAction.ADD_AFTER:
            if self.parent is None:
                raise ValueError("Cannot move after without parent")
            index = self.parent.index(self)
            self.parent.insert(index + 1, node)
        elif add_action == AddAction.ADD_TO_HEAD:
            cast(Group, self).prepend(node)
        elif add_action == AddAction.ADD_TO_TAIL:
            cast(Group, self).append(node)
        elif add_action == AddAction.REPLACE:
            self.replace_with(node)
        return node

    def pause(self):
        import supriya.commands

        if self.is_paused:
            return
        self._is_paused = True
        if self.is_allocated:
            request = supriya.commands.NodeRunRequest(
                node_id_run_flag_pairs=((self.node_id, False),)
            )
            request.communicate(server=self.server, sync=False)

    def precede_by(self, expr):
        if not isinstance(expr, Sequence):
            expr = [expr]
        index = self.parent.index(self)
        self.parent[index:index] = expr

    def query(self):
        from supriya.commands import NodeQueryRequest
        from supriya.querytree import QueryTreeGroup, QueryTreeSynth

        query_tree = {}
        stack = [self.node_id]
        while stack:
            node_id = stack.pop()
            if node_id in query_tree:
                continue
            request = NodeQueryRequest(node_id)
            response = request.communicate(server=self.server)
            if (response.next_node_id or -1) > 0:
                stack.append(response.next_node_id)
            if (response.head_node_id or -1) > 0:
                stack.append(response.head_node_id)
            if response.is_group:
                query_tree[node_id] = QueryTreeGroup.from_response(response)
            else:
                query_tree[node_id] = QueryTreeSynth.from_response(response)
            if response.parent_id in query_tree:
                query_tree[response.parent_id]._children += (query_tree[node_id],)
        return query_tree[self.node_id]

    def replace_with(self, expr):
        if not isinstance(expr, Sequence):
            expr = [expr]
        index = self.parent.index(self)
        self.parent[index : index + 1] = expr

    def succeed_by(self, expr):
        if not isinstance(expr, Sequence):
            expr = [expr]
        index = self.parent.index(self)
        self.parent[index + 1 : index + 1] = expr

    def unpause(self):
        import supriya.commands

        if not self.is_paused:
            return
        self._is_paused = False
        if self.is_allocated:
            request = supriya.commands.NodeRunRequest(
                node_id_run_flag_pairs=((self.node_id, True),)
            )
            request.communicate(server=self.server, sync=False)

    ### PUBLIC PROPERTIES ###

    @property
    def is_allocated(self) -> bool:
        if self.server is not None:
            return self in self.server
        return False

    @property
    def is_paused(self) -> bool:
        return self._is_paused

    @property
    def node_id(self) -> Optional[int]:
        return self._node_id

    @property
    def node_id_is_permanent(self) -> bool:
        return self._node_id_is_permanent

    @property
    def server(self) -> Optional["supriya.realtime.servers.Server"]:
        return self._server


class Group(Node, UniqueTreeList):
    """
    A group.

    ::

        >>> import supriya.realtime
        >>> server = supriya.Server()
        >>> server.boot()
        <Server: udp://127.0.0.1:57110, 8i8o>

    ::

        >>> group = supriya.realtime.Group()
        >>> group.allocate(server)
        <+ Group: 1000>

    ::

        >>> group.free()
        <- Group: ???>

    ::

        >>> server.quit()
        <Server: offline>

    """

    ### CLASS VARIABLES ###

    _valid_add_actions: Tuple[int, ...] = (
        AddAction.ADD_TO_HEAD,
        AddAction.ADD_TO_TAIL,
        AddAction.ADD_AFTER,
        AddAction.ADD_BEFORE,
        AddAction.REPLACE,
    )

    ### INITIALIZER ###

    def __init__(self, children=None, name=None, node_id_is_permanent=False):
        self._control_interface = GroupInterface(client=self)
        Node.__init__(self, name=name, node_id_is_permanent=node_id_is_permanent)
        UniqueTreeList.__init__(self, children=children, name=name)

    ### SPECIAL METHODS ###

    def __graph__(self):
        graph = super().__graph__()
        parent_node = graph[self._get_graphviz_name()]
        for child in self:
            graph.extend(child.__graph__())
            child_node = graph[child._get_graphviz_name()]
            parent_node.attach(child_node)
        return graph

    def __setitem__(self, i, expr):
        """
        Sets `expr` in self at index `i`.

        ::

            >>> group_one = supriya.realtime.Group()
            >>> group_two = supriya.realtime.Group()
            >>> group_one.append(group_two)

        """
        # TODO: lean on uqbar's __setitem__ more.
        self._validate(expr)
        if isinstance(i, slice):
            assert isinstance(expr, Sequence)
        if isinstance(i, str):
            i = self.index(self._named_children[i])
        if isinstance(i, int):
            if i < 0:
                i = len(self) + i
            i = slice(i, i + 1)
        if (
            i.start == i.stop
            and i.start is not None
            and i.stop is not None
            and i.start <= -len(self)
        ):
            start, stop = 0, 0
        else:
            start, stop, stride = i.indices(len(self))
        if not isinstance(expr, Sequence):
            expr = [expr]
        if self.is_allocated:
            self._set_allocated(expr, start, stop)
        else:
            self._set_unallocated(expr, start, stop)

    def __str__(self):
        result = []
        node_id = self.node_id
        if node_id is None:
            node_id = "???"
        if self.name:
            string = f"{node_id} group ({self.name})"
        else:
            string = f"{node_id} group"
        result.append(string)
        for child in self:
            lines = str(child).splitlines()
            for line in lines:
                result.append(f"    {line}")
        return "\n".join(result)

    ### PRIVATE METHODS ###

    def _as_graphviz_node(self):
        node = super()._as_graphviz_node()
        node.attributes["fillcolor"] = "lightsteelblue2"
        return node

    @staticmethod
    def _iterate_setitem_expr(group, expr, start=0):
        if not start or not group:
            outer_target_node = group
        else:
            outer_target_node = group[start - 1]
        for outer_node in expr:
            if outer_target_node is group:
                outer_add_action = AddAction.ADD_TO_HEAD
            else:
                outer_add_action = AddAction.ADD_AFTER
            outer_node_was_allocated = outer_node.is_allocated
            yield outer_node, outer_target_node, outer_add_action
            outer_target_node = outer_node
            if isinstance(outer_node, Group) and not outer_node_was_allocated:
                for (
                    inner_node,
                    inner_target_node,
                    inner_add_action,
                ) in Group._iterate_setitem_expr(outer_node, outer_node):
                    yield inner_node, inner_target_node, inner_add_action

    def _collect_requests_and_synthdefs(self, expr, server, start=0):
        import supriya.commands

        nodes = set()
        paused_nodes = set()
        synthdefs = set()
        requests = []
        iterator = Group._iterate_setitem_expr(self, expr, start)
        for node, target_node, add_action in iterator:
            nodes.add(node)
            if node.is_allocated:
                if add_action == AddAction.ADD_TO_HEAD:
                    request = supriya.commands.GroupHeadRequest(
                        node_id_pairs=[(node, target_node)]
                    )
                else:
                    request = supriya.commands.NodeAfterRequest(
                        node_id_pairs=[(node, target_node)]
                    )
                requests.append(request)
            else:
                if isinstance(node, Group):
                    request = supriya.commands.GroupNewRequest(
                        items=[
                            supriya.commands.GroupNewRequest.Item(
                                add_action=add_action,
                                node_id=node,
                                target_node_id=target_node,
                            )
                        ]
                    )
                    requests.append(request)
                else:
                    if node.synthdef not in server:
                        synthdefs.add(node.synthdef)
                    (settings, map_requests) = node.controls._make_synth_new_settings()
                    request = supriya.commands.SynthNewRequest(
                        add_action=add_action,
                        node_id=node,
                        synthdef=node.synthdef,
                        target_node_id=target_node,
                        **settings,
                    )
                    requests.append(request)
                    requests.extend(map_requests)
                if node.is_paused:
                    paused_nodes.add(node)
        return nodes, paused_nodes, requests, synthdefs

    def _set_allocated(self, expr, start, stop):
        # TODO: Consolidate this with Group.allocate()
        # TODO: Perform tree mutations via command apply methods, not here
        import supriya.commands

        old_nodes = self._children[start:stop]
        self._children.__delitem__(slice(start, stop))
        for old_node in old_nodes:
            old_node._set_parent(None)
        for child in expr:
            if child in self and self.index(child) < start:
                start -= 1
            child._set_parent(self)
        self._children.__setitem__(slice(start, start), expr)
        (
            new_nodes,
            paused_nodes,
            requests,
            synthdefs,
        ) = self._collect_requests_and_synthdefs(expr, self.server, start=start)
        nodes_to_free = [_ for _ in old_nodes if _ not in new_nodes]
        if nodes_to_free:
            requests.append(
                supriya.commands.NodeFreeRequest(
                    node_ids=sorted(nodes_to_free, key=lambda x: x.node_id)
                )
            )
        return self._allocate(paused_nodes, requests, self.server, synthdefs)

    def _set_unallocated(self, expr, start, stop):
        for node in expr:
            node.free()
        for old_child in tuple(self[start:stop]):
            old_child._set_parent(None)
        self._children[start:stop] = expr
        for new_child in expr:
            new_child._set_parent(self)

    def _unregister_with_local_server(self):
        for child in self:
            child._unregister_with_local_server()
        return Node._unregister_with_local_server(self)

    def _validate(self, expr):
        assert all(isinstance(_, supriya.realtime.Node) for _ in expr)
        parentage = self.parentage
        for x in expr:
            assert isinstance(x, Node)
            if isinstance(x, Group):
                assert x not in parentage

    ### PUBLIC METHODS ###

    def allocate(
        self, target_node, add_action=None, node_id_is_permanent=False, sync=False
    ):
        # TODO: Consolidate this with Group.allocate()
        import supriya.commands

        if self.is_allocated:
            return
        self._node_id_is_permanent = bool(node_id_is_permanent)
        target_node = Node._expr_as_target(target_node)
        server = target_node.server
        group_new_request = supriya.commands.GroupNewRequest(
            items=[
                supriya.commands.GroupNewRequest.Item(
                    add_action=add_action,
                    node_id=self,
                    target_node_id=target_node.node_id,
                )
            ]
        )
        (
            nodes,
            paused_nodes,
            requests,
            synthdefs,
        ) = self._collect_requests_and_synthdefs(self, server)
        requests = [group_new_request, *requests]
        if self.is_paused:
            paused_nodes.add(self)
        return self._allocate(paused_nodes, requests, server, synthdefs)

    def free(self):
        for node in self:
            node._unregister_with_local_server()
        Node.free(self)
        return self

    def prepend(self, expr: Node):
        self[0:0] = [expr]

    ### PUBLIC PROPERTIES ###

    @property
    def controls(self) -> GroupInterface:
        return self._control_interface


class Synth(Node):
    """
    A synth.

    ::

        >>> import supriya.realtime
        >>> server = supriya.Server()
        >>> server.boot()
        <Server: udp://127.0.0.1:57110, 8i8o>

    ::

        >>> import supriya.synthdefs
        >>> import supriya.ugens
        >>> with supriya.synthdefs.SynthDefBuilder(
        ...     amplitude=0.0,
        ...     frequency=440.0,
        ... ) as builder:
        ...     sin_osc = supriya.ugens.SinOsc.ar(
        ...         frequency=builder["frequency"],
        ...     )
        ...     sin_osc *= builder["amplitude"]
        ...     out = supriya.ugens.Out.ar(
        ...         bus=0,
        ...         source=[sin_osc, sin_osc],
        ...     )
        ...
        >>> synthdef = builder.build()
        >>> synthdef.allocate(server)
        <SynthDef: e41193ac8b7216f49ff0d477876a3bf3>

    ::

        >>> synth = supriya.realtime.Synth(amplitude=0.5, frequency=443, synthdef=synthdef)
        >>> synth
        <- Synth: ??? e41193ac8b7216f49ff0d477876a3bf3>

    ::

        >>> synth.allocate(server)
        <+ Synth: 1000 e41193ac8b7216f49ff0d477876a3bf3>

    ::

        >>> synth["frequency"] = 666.0
        >>> synth["frequency"]
        666.0

    ::

        >>> server.quit()
        <Server: offline>

    """

    ### CLASS VARIABLES ###

    _valid_add_actions = (AddAction.ADD_BEFORE, AddAction.ADD_AFTER, AddAction.REPLACE)

    ### INITIALIZER ###

    def __init__(
        self,
        synthdef: Optional[SynthDef] = None,
        name=None,
        node_id_is_permanent=False,
        **kwargs,
    ):
        synthdef = synthdef or supriya.assets.synthdefs.default
        if not isinstance(synthdef, SynthDef):
            raise ValueError(synthdef)
        Node.__init__(self, name=name, node_id_is_permanent=node_id_is_permanent)
        self._synthdef = synthdef
        self._control_interface = SynthInterface(client=self, synthdef=self._synthdef)
        self._control_interface._set(**kwargs)

    ### SPECIAL METHODS ###

    def __getitem__(self, item):
        return self._control_interface[item].value

    def __iter__(self):
        return iter(self._control_interface)

    def __repr__(self):
        class_name = type(self).__name__
        node_id = "???"
        if self.node_id is not None:
            node_id = self.node_id
        allocated = "+" if self.is_allocated else "-"
        synthdef_name = "???"
        if self.synthdef is not None:
            synthdef_name = self.synthdef.actual_name
        if self.name is None:
            return f"<{allocated} {class_name}: {node_id} {synthdef_name}>"
        return f"<{allocated} {class_name}: {node_id} {synthdef_name} ({self.name})>"

    def __setitem__(self, items, values):
        self.controls.__setitem__(items, values)

    def __str__(self):
        result = []
        node_id = self.node_id
        if node_id is None:
            node_id = "???"
        if self.name:
            string = "{node_id} {synthdef} ({name})"
        else:
            string = "{node_id} {synthdef}"
        string = string.format(
            name=self.name, node_id=node_id, synthdef=self.synthdef.actual_name
        )
        result.append(string)
        control_pieces = []
        for _, parameter in sorted(self.synthdef.indexed_parameters):
            control = self.controls[parameter.name]
            control_piece = "{}: {!s}".format(control.name, control.value)
            control_pieces.append(control_piece)
        control_pieces = "    " + ", ".join(control_pieces)
        result.append(control_pieces)
        result = "\n".join(result)
        return result

    ### PRIVATE METHODS ###

    def _as_graphviz_node(self):
        node = super()._as_graphviz_node()
        node.attributes["fillcolor"] = "lightgoldenrod2"
        return node

    def _unregister_with_local_server(self):
        node_id = Node._unregister_with_local_server(self)
        if "gate" in self.controls:
            self.controls["gate"].reset()
        return node_id

    ### PUBLIC METHODS ###

    def allocate(
        self,
        target_node,
        add_action=None,
        node_id_is_permanent=False,
        sync=True,
        **kwargs,
    ):
        import supriya.commands

        if self.is_allocated:
            return
        self._node_id_is_permanent = bool(node_id_is_permanent)
        target_node = Node._expr_as_target(target_node)
        server = target_node.server
        if not server.is_running:
            raise ServerOffline
        self.controls._set(**kwargs)
        # TODO: Map requests aren't necessary during /s_new
        settings, map_requests = self.controls._make_synth_new_settings()
        synth_request = supriya.commands.SynthNewRequest(
            add_action=add_action,
            node_id=self,
            synthdef=self.synthdef,
            target_node_id=target_node.node_id,
            **settings,
        )
        requests = [synth_request, *map_requests]
        paused_nodes = set()
        synthdefs = set()
        if self.is_paused:
            paused_nodes.add(self)
        if self.synthdef not in server:
            synthdefs.add(self.synthdef)
        return self._allocate(paused_nodes, requests, server, synthdefs)

    def release(self):
        if self.is_allocated and "gate" in self.controls:
            self["gate"] = 0
        else:
            self.free()
        return self

    ### PUBLIC PROPERTIES ###

    @property
    def controls(self) -> SynthInterface:
        return self._control_interface

    @property
    def synthdef(self) -> SynthDef:
        return self._synthdef


class RootNode(Group):

    ### CLASS VARIABLES ###

    _valid_add_actions: Tuple[int, ...] = (AddAction.ADD_TO_HEAD, AddAction.ADD_TO_TAIL)

    ### INITIALIZER ###

    def __init__(self, server):
        super().__init__()
        self._server = server

    ### SPECIAL METHODS ###

    def __str__(self):
        return "NODE TREE " + super().__str__()

    ### PRIVATE METHODS ###

    def _as_graphviz_node(self):
        node = super()._as_graphviz_node()
        node.attributes["fillcolor"] = "lightsalmon2"
        return node

    ### PUBLIC METHODS ###

    def allocate(self):
        pass

    def free(self):
        pass

    def pause(self):
        pass

    def unpause(self):
        pass

    ### PUBLIC PROPERTIES ###

    @property
    def node_id(self) -> int:
        return 0

    @property
    def parent(self) -> None:
        return None



"""
Tools for working in non-realtime.
"""
from .bases import SessionObject
from .buffers import Buffer, BufferGroup
from .buses import AudioInputBusGroup, AudioOutputBusGroup, Bus, BusGroup
from .nodes import Group, Node, RootNode, Synth
from .renderer import SessionRenderer
from .sessions import Session
from .states import DoNotPropagate, Moment, NodeTransition, State

__all__ = [
    "AudioInputBusGroup",
    "AudioOutputBusGroup",
    "Buffer",
    "BufferGroup",
    "Bus",
    "BusGroup",
    "DoNotPropagate",
    "Group",
    "Moment",
    "Node",
    "NodeTransition",
    "RootNode",
    "Session",
    "SessionObject",
    "SessionRenderer",
    "State",
    "Synth",
]



import os
import pathlib
import platform

import uqbar.io

import supriya


def find(sclang_path=None):
    """Find the ``sclang`` executable.

    The following paths, if defined, will be searched (prioritised as ordered):
    1. The absolute path ``sclang_path``
    2. The environment variable ``SCLANG_PATH``
    3. ``sclang_path`` if defined in Supriya's configuration file
    4. The user's ``PATH``
    5. Common installation directories of the SuperCollider application.

    Returns a path to the ``sclang`` executable.
    Raises ``RuntimeError`` if no path is found.
    """
    sclang_path = pathlib.Path(
        sclang_path
        or os.environ.get("SCLANG_PATH")
        or supriya.config.get("core", "sclang_path", fallback=None)
        or "sclang"
    )
    if sclang_path.is_absolute() and uqbar.io.find_executable(sclang_path):
        return sclang_path
    sclang_path_candidates = uqbar.io.find_executable(sclang_path.name)
    if sclang_path_candidates:
        return pathlib.Path(sclang_path_candidates[0])
    if platform.system() == "Darwin":
        for path in [
            pathlib.Path("/Applications/SuperCollider.app/Contents/MacOS/sclang"),
            pathlib.Path(
                "/Applications/SuperCollider/SuperCollider.app/Contents/MacOS/sclang"
            ),
        ]:
            if path.exists():
                return path
    elif platform.system() == "Linux":
        for path in [
            pathlib.Path("/usr/bin/sclang"),
            pathlib.Path("/usr/local/bin/sclang"),
        ]:
            if path.exists():
                return path
    raise RuntimeError("Failed to locate sclang")
